from typing import Any, Optional, List

from sirius_sdk.agent.connections import AgentRPC
from sirius_sdk.agent.wallet.abstract.did import AbstractDID


class DIDProxy(AbstractDID):

    def __init__(self, rpc: AgentRPC):
        self.__rpc = rpc

    async def create_and_store_my_did(self, did: str = None, seed: str = None, cid: bool = None) -> (str, str):
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/create_and_store_my_did',
            params=dict(did=did, seed=seed, cid=cid)
        )

    async def store_their_did(self, did: str, verkey: str = None) -> None:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/store_their_did',
            params=dict(did=did, verkey=verkey)
        )

    async def set_did_metadata(self, did: str, metadata: dict = None) -> None:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/set_did_metadata',
            params=dict(did=did, metadata=metadata)
        )

    async def list_my_dids_with_meta(self) -> List[Any]:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/list_my_dids_with_meta'
        )

    async def get_did_metadata(self, did) -> Optional[dict]:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/get_did_metadata',
            params=dict(did=did)
        )

    async def key_for_local_did(self, did: str) -> str:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/key_for_local_did',
            params=dict(did=did)
        )

    async def key_for_did(self, pool_name: str, did: str) -> str:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/key_for_did',
            params=dict(pool_name=pool_name, did=did)
        )

    async def create_key(self, seed: str = None) -> str:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/create_key__did',
            params=dict(seed=seed)
        )

    async def replace_keys_start(self, did: str, seed: str = None) -> str:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/replace_keys_start',
            params=dict(did=did, seed=seed)
        )

    async def replace_keys_apply(self, did: str) -> None:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/replace_keys_apply',
            params=dict(did=did)
        )

    async def set_key_metadata(self, verkey: str, metadata: dict) -> None:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/set_key_metadata__did',
            params=dict(verkey=verkey, metadata=metadata)
        )

    async def get_key_metadata(self, verkey: str) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/get_key_metadata__did',
            params=dict(verkey=verkey)
        )

    async def set_endpoint_for_did(self, did: str, address: str, transport_key: str) -> None:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/set_endpoint_for_did',
            params=dict(did=did, address=address, transport_key=transport_key)
        )

    async def get_endpoint_for_did(self, pool_name: str, did: str) -> (str, Optional[str]):
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/get_endpoint_for_did',
            params=dict(pool_name=pool_name, did=did)
        )

    async def get_my_did_with_meta(self, did: str) -> Any:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/get_my_did_with_meta',
            params=dict(did=did)
        )

    async def abbreviate_verkey(self, did: str, full_verkey: str) -> str:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/abbreviate_verkey',
            params=dict(did=did, full_verkey=full_verkey)
        )

    async def qualify_did(self, did: str, method: str) -> str:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/qualify_did',
            params=dict(did=did, method=method)
        )



from .messages import PickUpBatchRequest, PickUpBatchResponse, PickUpListRequest, PickUpListResponse, \
    PickUpStatusRequest, PickUpStatusResponse, PickUpProblemReport, PickUpNoop


__all__ = [
    "PickUpBatchRequest", "PickUpBatchResponse", "PickUpListRequest", "PickUpListResponse",
    "PickUpStatusRequest", "PickUpStatusResponse", "PickUpProblemReport", "PickUpNoop"
]



import asyncio
import uuid

import pytest

from sirius_sdk.errors.exceptions import OperationAbortedManually
from sirius_sdk.errors.exceptions import SiriusTimeoutIO
from sirius_sdk.hub.defaults.inmemory_bus import InMemoryBus


@pytest.mark.asyncio
async def test_sane():
    session1 = InMemoryBus()
    session2 = InMemoryBus()
    thid1 = 'thread-' + uuid.uuid4().hex
    thid2 = 'thread-' + uuid.uuid4().hex
    content1 = b'Message-1'
    content2 = b'Message-2'

    # Subscribe from session-2
    for thid in [thid1, thid2]:
        ok = await session2.subscribe(thid)
        assert ok is True
    # Publish from session-1
    for thid, content in [(thid1, content1), (thid2, content2)]:
        num = await session1.publish(thid, content)
        assert num > 0
    # Retrieve from session-2
    for n in range(2):
        event = await session2.get_event(timeout=3)
        assert event.payload in [content1, content2]
    # Unsubscribe from thread-2
    await session2.unsubscribe(thid1)
    # Publish again
    for thid, num_expected in [(thid1, 0), (thid2, 1)]:
        num = await session1.publish(thid, content)
        assert num == num_expected
    # Retrieve from session-2
    event = await session2.get_event(timeout=3)
    assert event.payload == content2
    with pytest.raises(SiriusTimeoutIO):
        await session2.get_event(timeout=3)


@pytest.mark.asyncio
async def test_abort(mediator_invitation: dict):
    session = InMemoryBus()
    thid = 'thread-id-' + uuid.uuid4().hex

    ok = await session.subscribe(thid)
    assert ok is True

    async def __abort():
        await asyncio.sleep(1)
        await session.abort()

    asyncio.ensure_future(__abort())
    with pytest.raises(OperationAbortedManually):
        await session.get_event()



import uuid
import asyncio

import pytest

import sirius_sdk
import sirius_sdk.hub
from sirius_sdk import Agent, TheirEndpoint, Pairwise
import sirius_sdk.hub.coprotocols
from sirius_sdk.errors.exceptions import *
from sirius_sdk.messaging import Message
from tests.conftest import get_pairwise, get_pairwise3
from tests.helpers import run_coroutines
from tests.helpers import ServerTestSuite

from .helpers import MSG_LOG, check_msg_log, TEST_MSG_TYPES, check_thread_orders


async def routine1(
        co: sirius_sdk.hub.coprotocols.AbstractP2PCoProtocol,
        server_address: str = None, credentials: bytes = None, p2p: sirius_sdk.P2PConnection = None,
        **kwargs
):
    server_address = server_address or kwargs.get('server_uri')
    async with sirius_sdk.context(server_uri=server_address, credentials=credentials, p2p=p2p):
        first_req = Message({
            '@type': TEST_MSG_TYPES[0],
            'content': 'Request1'
        })
        MSG_LOG.append(first_req)
        print('#2')
        ok, resp1 = await co.switch(message=first_req)
        print('#2')
        assert ok is True
        MSG_LOG.append(Message(**resp1))
        ok, resp2 = await co.switch(
            message=Message({
                '@type': TEST_MSG_TYPES[2],
                'content': 'Request2'
            })
        )
        assert ok is True
        MSG_LOG.append(resp2)


async def routine2(
        co: sirius_sdk.hub.coprotocols.AbstractP2PCoProtocol,
        server_address: str = None, credentials: bytes = None, p2p: sirius_sdk.P2PConnection = None,
        **kwargs
):
    server_address = server_address or kwargs.get('server_uri')
    async with sirius_sdk.context(server_uri=server_address, credentials=credentials, p2p=p2p):
        await asyncio.sleep(1)
        print('#1')
        ok, resp1 = await co.switch(
            message=Message({
                '@type': TEST_MSG_TYPES[1],
                'content': 'Response1'
            })
        )
        print('#2')
        assert ok is True
        MSG_LOG.append(Message(**resp1))
        await co.send(
            message=Message({
                '@type': TEST_MSG_TYPES[3],
                'content': 'End'
            })
        )


@pytest.mark.asyncio
async def test__their_endpoint_coprotocol(test_suite: ServerTestSuite):
    agent1_params = test_suite.get_agent_params('agent1')
    agent2_params = test_suite.get_agent_params('agent2')
    entity1 = list(agent1_params['entities'].items())[0][1]
    entity2 = list(agent2_params['entities'].items())[0][1]

    async with sirius_sdk.context(**agent1_params):
        agent1_endpoint = [e for e in await sirius_sdk.endpoints() if e.routing_keys == []][0].address
    async with sirius_sdk.context(**agent2_params):
        agent2_endpoint = [e for e in await sirius_sdk.endpoints() if e.routing_keys == []][0].address

    # FIRE!!!
    their1 = TheirEndpoint(agent2_endpoint, entity2['verkey'])
    their2 = TheirEndpoint(agent1_endpoint, entity1['verkey'])
    co1 = sirius_sdk.hub.coprotocols.CoProtocolP2PAnon(entity1['verkey'], their1, ['test_protocol'])
    co2 = sirius_sdk.hub.coprotocols.CoProtocolP2PAnon(entity2['verkey'], their2, ['test_protocol'])
    MSG_LOG.clear()
    await run_coroutines(
        routine1(co1, **agent1_params),
        routine2(co2, **agent2_params),
        timeout=5
    )
    check_msg_log()


@pytest.mark.asyncio
async def test__pairwise_coprotocol(config_a: dict, config_b: dict):
    a2b = await get_pairwise3(me=config_a, their=config_b)
    b2a = await get_pairwise3(me=config_b, their=config_a)

    co1 = sirius_sdk.hub.coprotocols.CoProtocolP2P(a2b, ['test_protocol'])
    co2 = sirius_sdk.hub.coprotocols.CoProtocolP2P(b2a, ['test_protocol'])
    MSG_LOG.clear()
    await run_coroutines(
        routine1(co1, **config_a),
        routine2(co2, **config_b)
    )
    check_msg_log()


@pytest.mark.asyncio
async def test__threadbased_coprotocol(config_a: dict, config_b: dict):
    a2b = await get_pairwise3(me=config_a, their=config_b)
    b2a = await get_pairwise3(me=config_b, their=config_a)
    thread_id = uuid.uuid4().hex
    co1 = sirius_sdk.hub.coprotocols.CoProtocolThreadedP2P(thread_id, a2b)
    co2 = sirius_sdk.hub.coprotocols.CoProtocolThreadedP2P(thread_id, b2a)
    MSG_LOG.clear()
    await run_coroutines(
        routine1(co1, **config_a),
        routine2(co2, **config_b),
    )
    check_msg_log()
    check_thread_orders()


@pytest.mark.asyncio
async def test__coprotocols_intersections(config_a: dict, config_b: dict):
    a2b = await get_pairwise3(me=config_a, their=config_b)
    b2a = await get_pairwise3(me=config_b, their=config_a)
    thread_id = uuid.uuid4().hex
    co1_threaded = sirius_sdk.hub.coprotocols.CoProtocolThreadedP2P(thread_id, a2b)
    co2_threaded = sirius_sdk.hub.coprotocols.CoProtocolThreadedP2P(thread_id, b2a)
    co1_p2p = sirius_sdk.hub.coprotocols.CoProtocolP2P(a2b, ['test_protocol'])
    co2_p2p = sirius_sdk.hub.coprotocols.CoProtocolP2P(b2a, ['test_protocol'])
    try:
        MSG_LOG.clear()
        await run_coroutines(
            routine1(co1_threaded, **config_a), routine2(co2_threaded, **config_b),
            routine1(co1_p2p, **config_a), routine2(co2_p2p, **config_b),
        )
        # collect messages
        threaded_sequence = [msg for msg in MSG_LOG if '~thread' in msg]
        non_threaded_sequence = [msg for msg in MSG_LOG if '~thread' not in msg]
        # threaded messages
        MSG_LOG.clear()
        MSG_LOG.extend(threaded_sequence)
        check_msg_log()
        for msg in threaded_sequence:
            assert msg['~thread']['thid'] == thread_id
        # non-threaded messages
        MSG_LOG.clear()
        MSG_LOG.extend(non_threaded_sequence)
        check_msg_log()
    finally:
        await co1_threaded.stop()
        await co2_threaded.stop()
        await co1_p2p.stop()
        await co2_p2p.stop()


@pytest.mark.asyncio
async def test_coprotocol_abort(config_a: dict, config_b: dict):
    a2b = await get_pairwise3(config_a, config_b)
    co = sirius_sdk.hub.coprotocols.CoProtocolThreadedP2P(thid=uuid.uuid4().hex, to=a2b, time_to_live=100)
    assert co.is_alive is False
    assert co.is_aborted is False
    exc = None

    async def infinite_reader(**kwargs):
        nonlocal co
        nonlocal exc
        try:
            async with sirius_sdk.context(**kwargs):
                while True:
                    msg = await co.get_one()
                    print(str(msg))
        except OperationAbortedManually as e:
            exc = e

    async def delayed_aborter():
        nonlocal co
        # Wait for coprotocol will be started implicitly
        await asyncio.sleep(1)
        assert co.is_alive is True
        assert co.is_aborted is False
        await co.abort()

    await run_coroutines(
        infinite_reader(**config_a),
        delayed_aborter(),
    )
    assert exc is not None
    assert isinstance(exc, OperationAbortedManually)
    assert co.is_aborted is True
    assert co.is_alive is False


@pytest.mark.asyncio
async def test_coprotocol_abort_multiple_ops_single_hub(config_a: dict, config_b: dict):
    a2b = await get_pairwise3(config_a, config_b)
    b2a = await get_pairwise3(config_b, config_a)

    co = sirius_sdk.hub.coprotocols.CoProtocolThreadedP2P(thid=uuid.uuid4().hex, to=a2b)
    new_thread_id = 'new-thread-id-' + uuid.uuid4().hex
    msg_log = []

    async def coroutine1(**kwargs):
        nonlocal co
        nonlocal msg_log
        nonlocal new_thread_id
        async with sirius_sdk.context(**kwargs):
            try:
                while True:
                    msg = await co.get_one()
                    print(str(msg))
            except OperationAbortedManually:
                pass

            try:
                new_co_on_same_hub = sirius_sdk.hub.coprotocols.CoProtocolThreadedP2P(thid=new_thread_id, to=a2b)
                msg, _, _ = await new_co_on_same_hub.get_one()
                print('!')
                msg_log.append(msg)
            except Exception as e:
                raise

    async def coroutine2(**kwargs):
        nonlocal co
        nonlocal new_thread_id
        await asyncio.sleep(3)

        await co.abort()

        await asyncio.sleep(3)
        async with sirius_sdk.context(**kwargs):
            try:
                new_co_on_same_hub = sirius_sdk.hub.coprotocols.CoProtocolThreadedP2P(thid=new_thread_id, to=b2a)
                await new_co_on_same_hub.send(
                    sirius_sdk.aries_rfc.Ping(comment='Test Ping')
                )
            except Exception as e:
                raise

    await run_coroutines(
        coroutine1(**config_a),
        coroutine2(**config_b)
    )
    assert len(msg_log) == 1
    msg = msg_log[0]
    assert isinstance(msg, sirius_sdk.aries_rfc.Ping)
    assert msg.comment == 'Test Ping'


@pytest.mark.asyncio
async def test_coprotocol_threaded_theirs_send(config_a: dict, config_b: dict, config_c: dict):
    a2b = await get_pairwise3(config_a, config_b)
    a2c = await get_pairwise3(config_a, config_c)
    thread_id = 'thread-id-' + uuid.uuid4().hex
    rcv_messages = []

    async def sender(**kwargs):
        nonlocal thread_id
        async with sirius_sdk.context(**kwargs):
            msg = sirius_sdk.aries_rfc.Ping(comment='Test Ping')
            co = sirius_sdk.hub.coprotocols.CoProtocolThreadedTheirs(thread_id, [a2b, a2c])
            await co.send(msg)

    async def reader(**kwargs):
        nonlocal rcv_messages
        async with sirius_sdk.context(**kwargs):
            listener = await sirius_sdk.subscribe()
            async for event in listener:
                rcv_messages.append(event.message)
                return

    await run_coroutines(
        sender(**config_a),
        reader(**config_b),
        reader(**config_c)
    )

    assert len(rcv_messages) == 2
    msg1 = rcv_messages[0]
    msg2 = rcv_messages[1]
    for msg in [msg1, msg2]:
        assert isinstance(msg1, sirius_sdk.aries_rfc.Ping)
        assert msg['~thread']['thid'] == thread_id
        assert msg.comment == 'Test Ping'


@pytest.mark.asyncio
async def test_coprotocol_threaded_theirs_switch(config_a: dict, config_b: dict, config_c: dict):
    a2b = await get_pairwise3(config_a, config_b)
    a2c = await get_pairwise3(config_a, config_c)

    thread_id = 'thread-id-' + uuid.uuid4().hex
    statuses = {}

    async def actor(**kwargs):
        nonlocal thread_id
        nonlocal statuses
        async with sirius_sdk.context(**kwargs):
            msg = sirius_sdk.aries_rfc.Ping(comment='Test Ping')
            co = sirius_sdk.hub.coprotocols.CoProtocolThreadedTheirs(thread_id, [a2b, a2c])
            statuses = await co.switch(msg)

    async def responder(**kwargs):
        async with sirius_sdk.context(**kwargs):
            listener = await sirius_sdk.subscribe()
            async for event in listener:
                thread_id_ = event.message['~thread']['thid']
                pong = sirius_sdk.aries_rfc.Pong(ping_id=thread_id_, comment='PONG')
                await sirius_sdk.send_to(pong, event.pairwise)
                return

    await run_coroutines(
        actor(**config_a),
        responder(**config_b),
        responder(**config_c),
    )

    assert statuses is not None
    assert a2b in statuses.keys()
    assert a2c in statuses.keys()
    for pw, stat in statuses.items():
        success, message = stat
        assert success is True
        assert message['comment'] == 'PONG'


@pytest.mark.asyncio
async def test_coprotocol_threaded_theirs_switch_timeout(config_a: dict, config_b: dict, config_c: dict):
    a2b = await get_pairwise3(config_a, config_b)
    a2c = await get_pairwise3(config_a, config_c)
    ttl = 10
    thread_id = 'thread-id-' + uuid.uuid4().hex
    statuses = {}

    async def actor(**kwargs):
        nonlocal thread_id
        nonlocal statuses
        nonlocal ttl
        async with sirius_sdk.context(**kwargs):
            msg = sirius_sdk.aries_rfc.Ping(comment='Test Ping')
            co = sirius_sdk.hub.coprotocols.CoProtocolThreadedTheirs(thread_id, [a2b, a2c], time_to_live=ttl)
            statuses = await co.switch(msg)

    async def responder(**kwargs):
        async with sirius_sdk.context(**kwargs):
            listener = await sirius_sdk.subscribe()
            async for event in listener:
                thread_id_ = event.message['~thread']['thid']
                pong = sirius_sdk.aries_rfc.Pong(ping_id=thread_id_, comment='PONG')
                await sirius_sdk.send_to(pong, event.pairwise)
                return

    await run_coroutines(
        actor(**config_a),
        responder(**config_b),
        timeout=2*ttl
    )

    assert statuses is not None
    stat1 = statuses[a2b]
    assert stat1[0] is True
    assert stat1[1]['comment'] == 'PONG'
    stat1 = statuses[a2c]
    assert stat1[0] is False
    assert stat1[1] is None


@pytest.mark.asyncio
async def test_coprotocol_timeouts(config_a: dict, config_b: dict):
    timeout = 1
    p2p = await get_pairwise3(me=config_a, their=config_b)

    co_under_test1 = sirius_sdk.hub.coprotocols.CoProtocolThreadedP2P(
        thid=uuid.uuid4().hex, to=p2p, time_to_live=timeout
    )
    co_under_test2 = sirius_sdk.hub.coprotocols.CoProtocolP2PAnon(
        p2p.me.verkey, p2p.their, ['test_protocol'], time_to_live=timeout
    )
    co_under_test3 = sirius_sdk.hub.coprotocols.CoProtocolP2P(
        p2p, ['test_protocol'], time_to_live=timeout
    )
    co_under_test4 = sirius_sdk.hub.coprotocols.CoProtocolThreadedTheirs(
        thid=uuid.uuid4().hex, theirs=[p2p]
    )

    async with sirius_sdk.context(**config_a):
        with pytest.raises(SiriusTimeoutIO):
            for co in [co_under_test1, co_under_test2, co_under_test3, co_under_test4]:
                await co.get_one()



import uuid

import pytest

import sirius_sdk

from sirius_sdk.hub.defaults.default_non_secrets import DefaultNonSecretsStorage
from sirius_sdk.agent.wallet.abstract.non_secrets import RetrieveRecordOptions
from sirius_sdk.hub.defaults.default_storage import InMemoryKeyValueStorage


@pytest.mark.asyncio
async def test_record_value_ops():
    obj_under_test = DefaultNonSecretsStorage(storage=InMemoryKeyValueStorage())
    type_under_test = 'type_' + uuid.uuid4().hex
    value = 'my-value'
    my_id = 'my-id-' + uuid.uuid4().hex
    await obj_under_test.add_wallet_record(type_under_test, my_id, value)
    opts = RetrieveRecordOptions()
    opts.check_all()
    value_info = await obj_under_test.get_wallet_record(type_under_test, my_id, opts)
    assert value_info['id'] == my_id
    assert value_info['tags'] == {}
    assert value_info['value'] == value
    assert value_info['type'] == type_under_test

    value_new = 'my-new-value'
    await obj_under_test.update_wallet_record_value(type_under_test, my_id, value_new)
    value_info = await obj_under_test.get_wallet_record(type_under_test, my_id, opts)
    assert value_info['value'] == value_new
    await obj_under_test.delete_wallet_record(type_under_test, my_id)


@pytest.mark.asyncio
async def test_record_tags_ops():
    obj_under_test = DefaultNonSecretsStorage(storage=InMemoryKeyValueStorage())
    type_under_test = 'type_' + uuid.uuid4().hex

    my_id = 'my-id-' + uuid.uuid4().hex

    value = 'my-value'
    tags = {
        'tag1': 'val1',
        '~tag2': 'val2'
    }
    await obj_under_test.add_wallet_record(type_under_test, my_id, value, tags)
    opts = RetrieveRecordOptions()
    opts.check_all()
    value_info = await obj_under_test.get_wallet_record(type_under_test, my_id, opts)
    assert value_info['id'] == my_id
    assert value_info['tags'] == tags
    assert value_info['value'] == value
    assert value_info['type'] == type_under_test

    upd_tags = {
        'ext-tag': 'val3'
    }
    await obj_under_test.update_wallet_record_tags(type_under_test, my_id, upd_tags)
    value_info = await obj_under_test.get_wallet_record(type_under_test, my_id, opts)
    assert value_info['tags'] == upd_tags

    await obj_under_test.add_wallet_record_tags(type_under_test, my_id, tags)
    value_info = await obj_under_test.get_wallet_record(type_under_test, my_id, opts)
    assert value_info['tags'] == dict(**upd_tags, **tags)

    await obj_under_test.delete_wallet_record_tags(type_under_test, my_id, ['ext-tag'])
    value_info = await obj_under_test.get_wallet_record(type_under_test, my_id, opts)
    assert value_info['tags'] == tags


@pytest.mark.asyncio
async def test_maintain_tags_only_update_ops():
    obj_under_test = DefaultNonSecretsStorage(storage=InMemoryKeyValueStorage())
    type_under_test = 'type_' + uuid.uuid4().hex

    my_id = 'my-id-' + uuid.uuid4().hex
    value = 'my-value'
    await obj_under_test.add_wallet_record(type_under_test, my_id, value, )
    opts = RetrieveRecordOptions()
    opts.check_all()
    value_info = await obj_under_test.get_wallet_record(type_under_test, my_id, opts)
    assert value_info['id'] == my_id
    assert value_info['tags'] == {}
    assert value_info['value'] == value
    assert value_info['type'] == type_under_test

    tags1 = {
        'tag1': 'val1',
        '~tag2': 'val2'
    }

    await obj_under_test.update_wallet_record_tags(type_under_test, my_id, tags1)
    value_info = await obj_under_test.get_wallet_record(type_under_test, my_id, opts)
    assert value_info['tags'] == tags1

    tags2 = {
        'tag3': 'val3',
    }
    await obj_under_test.update_wallet_record_tags(type_under_test, my_id, tags2)
    value_info = await obj_under_test.get_wallet_record(type_under_test, my_id, opts)
    assert value_info['tags'] == tags2


@pytest.mark.asyncio
async def test_wallet_search():
    obj_under_test = DefaultNonSecretsStorage(storage=InMemoryKeyValueStorage())

    my_id1 = 'my-id-' + uuid.uuid4().hex
    my_id2 = 'my-id-' + uuid.uuid4().hex
    type_ = 'type_' + uuid.uuid4().hex
    opts = RetrieveRecordOptions()
    opts.check_all()
    tags1 = {
        'tag1': 'val1',
        '~tag2': '5',
        'marker': 'A'
    }
    tags2 = {
        'tag3': 'val3',
        '~tag4': '6',
        'marker': 'B'
    }
    await obj_under_test.add_wallet_record(type_, my_id1, 'value1', tags1)
    await obj_under_test.add_wallet_record(type_, my_id2, 'value2', tags2)

    query = {
        "tag1": "val1"
    }
    records, total = await obj_under_test.wallet_search(type_, query, opts)
    assert total == 1
    assert 'value1' in str(records)

    query = {
        "tag1": "val1",
        "marker": "A"
    }
    records, total = await obj_under_test.wallet_search(type_, query, opts)
    assert total == 1
    assert 'value1' in str(records)

    query = {
        "$or": [{"tag1": "val1"}, {"~tag4": "6"}]
    }
    records, total = await obj_under_test.wallet_search(type_, query, opts)
    assert len(records) == 1
    assert total == 2

    records, total = await obj_under_test.wallet_search(type_, query, opts, limit=1000)
    assert len(records) == 2
    assert total == 2

    query = {
        "marker": {"$in": ["A", "C"]}
    }
    records, total = await obj_under_test.wallet_search(type_, query, opts)
    assert 1 == total



r"""This extension adds a new header in DIDComm messages: return_route. When a message is received with this header,
use of the connection to return messages should be adjusted according to the value presented.

Details: https://github.com/decentralized-identity/didcomm-messaging/blob/master/extensions/return_route/main.md#return-route-header
"""
import logging
from enum import Enum
from typing import Union, Optional


class RouteType(Enum):
    # Default. No messages should be returned over this connection. Default value.
    NONE = 'none'

    # Send all messages for this DID over the connection.
    ALL = 'all'

    # Send all messages matching the DID and thread specified in the return_route_thread attribute.
    THREAD = 'thread'


HEADER_NAME = 'return_route'


def extract_value(headers: Union[dict, list]) -> RouteType:
    """For HTTP transports, the presence of this message decorator indicates
    that the receiving agent MAY hold onto the connection and use it to return messages as designated.
    HTTP transports will only be able to receive at most one message at a time.
    Websocket transports are capable of receiving multiple messages over a single connection.

    :param headers: http headers of the transport
    """
    if type(headers) is dict:
        headers = list(headers.items())
    if type(headers) is not list:
        logging.warning('Unexpected headers type')
        return RouteType.NONE
    for item in headers:
        if type(item) is tuple and len(item) == 2:
            name, value = item
            if type(name) is bytes:
                name = name.decode()
            if name == HEADER_NAME:
                if type(value) is bytes:
                    value = value.decode()
                if value == RouteType.ALL.value:
                    return RouteType.ALL
                elif value == RouteType.THREAD.value:
                    return RouteType.THREAD
                else:
                    return RouteType.NONE
        else:
            logging.warning('Unexpected header type')
    return RouteType.NONE



import asyncio
import hashlib
import json
import logging
import warnings
import threading
from typing import Optional
from contextlib import asynccontextmanager

from sirius_sdk.errors.exceptions import SiriusInitializationError
from sirius_sdk.agent.pairwise import AbstractPairwiseList
from sirius_sdk.abstract.api import APICrypto, APICoProtocols, APITransport, APIContents, APIDistributedLocks,\
    APIRouter, APINetworks
from sirius_sdk.agent.wallet.abstract.cache import AbstractCache
from sirius_sdk.agent.wallet.abstract.did import AbstractDID
from sirius_sdk.agent.wallet.abstract.anoncreds import AbstractAnonCreds
from sirius_sdk.agent.wallet.abstract.non_secrets import AbstractNonSecrets
from sirius_sdk.abstract.storage import AbstractKeyValueStorage
from sirius_sdk.agent.microledgers.abstract import AbstractMicroledgerList
from sirius_sdk.agent.agent import Agent, SpawnStrategy

from .defaults.default_apis import APIDefault
from .defaults.default_crypto import DefaultCryptoService as DefaultCryptoService
from .defaults.default_storage import InMemoryKeyValueStorage
from .defaults.default_non_secrets import DefaultNonSecretsStorage
from .context import get as context_get, set as context_set, clear as context_clear
from .config import Config
from .mediator import Mediator
from .backgrounds import BackgroundScheduler

__ROOT_HUB = None
__THREAD_LOCAL_HUB = threading.local()


class Hub:

    def __init__(self, config: Config, loop: asyncio.AbstractEventLoop = None):

        self.__config: Config = config
        self.__loop = loop or asyncio.get_event_loop()
        self.__agent: Optional[Agent] = None
        self.__allocate_agent = False
        self.__mediator: Optional[Mediator] = None
        self.__allocate_mediator = False

        self.__storage: Optional[AbstractKeyValueStorage] = config.overrides.storage
        if self.__storage is None and not config.cloud_opts.is_filled:
            logging.warning(
                'Storage will be set to InMemory-Storage as default, it will outcome issues in production environments'
            )
            self.__storage = InMemoryKeyValueStorage()

        # Check if configured as cloud-agent
        if config.cloud_opts.is_filled:
            self.__allocate_agent = True
            self.__create_agent_instance(external_crypto=config.overrides.crypto)
        elif config.mediator_opts.is_filled:
            self.__allocate_mediator = True
            self.__create_mediator_instance(pairwise_resolver=config.overrides.pairwise_storage)
        else:
            logging.warning('You should configure cloud-agent or mediator options')
        # Crypto and default services
        self.__crypto: Optional[APICrypto] = config.overrides.crypto
        self.__default_api: APIDefault = APIDefault()
        self.__default_crypto: APICrypto = DefaultCryptoService(storage=self.__storage)
        self.__default_non_secrets = DefaultNonSecretsStorage(storage=self.__storage)

    def __del__(self):
        if self.__loop and self.__loop.is_running():
            if self.__agent is not None and self.__agent.is_open:
                asyncio.ensure_future(self.__agent.close(), loop=self.__loop)
            if self.__mediator is not None and self.__mediator.is_connected:
                asyncio.ensure_future(self.__mediator.disconnect(), loop=self.__loop)

    @property
    def global_id(self) -> Optional[str]:
        if self.__config.cloud_opts.is_filled:
            if isinstance(self.__config.cloud_opts.credentials, str):
                cred = self.__config.cloud_opts.credentials.encode()
            else:
                cred = self.__config.cloud_opts.credentials
            return hashlib.md5(cred).hexdigest()
        elif self.__config.mediator_opts.is_filled:
            cred = f'{self.__config.mediator_opts.my_verkey}:{self.__config.mediator_opts.mediator_verkey}'
            return hashlib.md5(cred).hexdigest()
        else:
            return None

    def copy(self):
        inst = Hub(config=self.__config)
        return inst

    async def abort(self):
        if not self.__allocate_agent:
            return
        if self.__loop.is_running():
            if self.__loop == asyncio.get_event_loop():
                old_agent = self.__agent
                self.__create_agent_instance(external_crypto=self.__config.overrides.crypto)
                if old_agent.is_open:
                    await old_agent.close()
            else:
                asyncio.ensure_future(self.abort(), loop=self.__loop)

    def run_soon(self, coro):
        assert asyncio.iscoroutine(coro), 'Expected coroutine object'
        if self.__loop.is_running():
            asyncio.ensure_future(coro, loop=self.__loop)

    @asynccontextmanager
    async def get_agent_connection_lazy(self):
        if self.__allocate_agent:
            if not self.__agent.is_open:
                await self.__agent.open()
            yield self.__agent
        elif self.__allocate_mediator:
            if not self.__mediator.is_connected:
                await self.__mediator.connect()
            yield self.__mediator
        else:
            yield None

    async def open(self):
        if not (self.__allocate_agent or self.__allocate_mediator):
            return
        async with self.get_agent_connection_lazy():
            pass

    async def close(self):
        if self.__allocate_agent and self.__agent and self.__agent.is_open:
            await self.__agent.close()
        if self.__allocate_mediator and self.__mediator and self.__mediator.is_connected:
            await self.__mediator.disconnect()

    async def get_crypto(self) -> APICrypto:
        if self.__allocate_agent:
            async with self.get_agent_connection_lazy() as agent:
                return self.__crypto or agent.wallet.crypto or self.__default_crypto
        else:
            return self.__crypto or self.__default_crypto

    async def get_microledgers(self) -> Optional[AbstractMicroledgerList]:
        if self.__allocate_agent:
            async with self.get_agent_connection_lazy() as agent:
                return self.__config.overrides.microledgers or agent.microledgers
        else:
            return self.__config.overrides.microledgers

    async def get_pairwise_list(self) -> Optional[AbstractPairwiseList]:
        if self.__allocate_agent:
            async with self.get_agent_connection_lazy() as agent:
                return self.__config.overrides.pairwise_storage or agent.pairwise_list
        else:
            return self.__config.overrides.pairwise_storage

    async def get_did(self) -> Optional[AbstractDID]:
        if self.__allocate_agent:
            async with self.get_agent_connection_lazy() as agent:
                return self.__config.overrides.did or agent.wallet.did
        else:
            return self.__config.overrides.did

    async def get_anoncreds(self) -> Optional[AbstractAnonCreds]:
        if self.__allocate_agent:
            async with self.get_agent_connection_lazy() as agent:
                return self.__config.overrides.anoncreds or agent.wallet.anoncreds
        else:
            return self.__config.overrides.anoncreds

    async def get_cache(self) -> Optional[AbstractCache]:
        if self.__allocate_agent:
            async with self.get_agent_connection_lazy() as agent:
                return self.__config.overrides.cache or agent.wallet.cache
        else:
            return self.__config.overrides.cache

    async def get_non_secrets(self) -> Optional[AbstractNonSecrets]:
        if self.__allocate_agent:
            async with self.get_agent_connection_lazy() as agent:
                return self.__config.overrides.non_secrets or agent.wallet.non_secrets
        else:
            return self.__config.overrides.non_secrets or self.__default_non_secrets

    async def get_coprotocols(self) -> Optional[APICoProtocols]:
        api: Optional[APICoProtocols] = None
        if self.__allocate_agent or self.__allocate_mediator:
            async with self.get_agent_connection_lazy() as conn:
                if isinstance(conn, APICoProtocols):
                    api = conn
        return self.__config.overrides.coprotocols or api or self.__default_api

    async def get_transport(self) -> Optional[APITransport]:
        api: Optional[APITransport] = None
        if self.__allocate_agent or self.__allocate_mediator:
            async with self.get_agent_connection_lazy() as conn:
                if isinstance(conn, APITransport):
                    api = conn
        return self.__config.overrides.coprotocols or api or self.__default_api

    async def get_contents(self) -> Optional[APIContents]:
        api: Optional[APIContents] = None
        if self.__allocate_agent or self.__allocate_mediator:
            async with self.get_agent_connection_lazy() as conn:
                if isinstance(conn, APIContents):
                    api = conn
        return self.__config.overrides.contents or api or self.__default_api

    async def get_distr_locks(self) -> Optional[APIDistributedLocks]:
        api: Optional[APIDistributedLocks] = None
        if self.__allocate_agent or self.__allocate_mediator:
            async with self.get_agent_connection_lazy() as conn:
                if isinstance(conn, APIDistributedLocks):
                    api = conn
        return self.__config.overrides.distr_locks or api or self.__default_api

    async def get_router(self) -> Optional[APIRouter]:
        api: Optional[APIRouter] = None
        if self.__allocate_agent or self.__allocate_mediator:
            async with self.get_agent_connection_lazy() as conn:
                if isinstance(conn, APIRouter):
                    api = conn
        return self.__config.overrides.router or api

    async def get_networks(self) -> Optional[APINetworks]:
        api: Optional[APINetworks] = None
        if self.__allocate_agent or self.__allocate_mediator:
            async with self.get_agent_connection_lazy() as conn:
                if isinstance(conn, APINetworks):
                    api = conn
        return self.__config.overrides.networks or api

    async def ping(self) -> bool:
        success = False
        if self.__allocate_agent:
            async with self.get_agent_connection_lazy() as agent:
                if isinstance(agent, Agent):
                    success = await agent.ping()
                elif isinstance(agent, Mediator):
                    success = agent.is_connected
        return success

    def __create_agent_instance(self, external_crypto: APICrypto):
        if self.__allocate_agent:
            self.__agent = Agent(
                server_address=self.__config.cloud_opts.server_uri,
                credentials=self.__config.cloud_opts.credentials,
                p2p=self.__config.cloud_opts.p2p,
                timeout=self.__config.cloud_opts.io_timeout,
                loop=self.__loop,
                storage=None,
                spawn_strategy=SpawnStrategy.CONCURRENT,
                external_crypto=external_crypto
            )

    def __create_mediator_instance(self, pairwise_resolver: AbstractPairwiseList = None):
        if self.__allocate_mediator:
            self.__mediator = Mediator(
                uri=self.__config.mediator_opts.uri,
                my_verkey=self.__config.mediator_opts.my_verkey,
                mediator_verkey=self.__config.mediator_opts.mediator_verkey,
                pairwise_resolver=pairwise_resolver
            )


def __restore_config_from_kwargs(*args, **kwargs) -> Config:
    if args:
        first_arg = args[0]
        if type(first_arg) is not Config:
            warnings.warn('Use sirius_sdk.Config to configure SDK', DeprecationWarning)
    if kwargs:
        for arg in kwargs.keys():
            warnings.warn(f'{arg} param is Deprecated. Use sirius_sdk.Config to configure SDK', DeprecationWarning)
    if args:
        first = args[0]
        if type(first) is Config:
            cfg = first
        else:
            cfg = Config()
        if len(args) >= 3:
            opts = list(args)[:3]
            cfg.setup_cloud(*opts)
    else:
        cfg = Config()

    server_uri = kwargs.pop('server_uri', None) or kwargs.pop('server_address', None)
    credentials = kwargs.pop('credentials', None)
    p2p = kwargs.pop('p2p', None)
    io_timeout = kwargs.pop('io_timeout', None)
    if server_uri and credentials and p2p:
        cfg.setup_cloud(server_uri=server_uri, credentials=credentials, p2p=p2p, io_timeout=io_timeout)

    cfg.override(**kwargs)
    return cfg


def init(cfg: Config = None, *args, **kwargs):

    if cfg is not None and not isinstance(cfg, Config):
        first_arg = cfg
        cfg = None
        args = tuple([first_arg] + list(args))
    if cfg is None:
        warnings.warn('Use sirius_sdk.Config to configure SDK', DeprecationWarning)
        cfg = __restore_config_from_kwargs(*args, **kwargs)

    global __ROOT_HUB
    root = Hub(cfg)
    loop = asyncio.get_event_loop()
    if loop.is_running():
        raise SiriusInitializationError('You must call this method outside coroutine')
    loop.run_until_complete(root.open())
    __ROOT_HUB = root


@asynccontextmanager
async def context(cfg: Config = None, *args, **kwargs):

    if cfg is not None and not isinstance(cfg, Config):
        first_arg = cfg
        cfg = None
        args = tuple([first_arg] + list(args))
    if cfg is None:
        warnings.warn('Use sirius_sdk.Config to configure SDK', DeprecationWarning)
        cfg = __restore_config_from_kwargs(*args, **kwargs)

    hub = Hub(cfg)
    old_hub = __get_thread_local_gub()
    __THREAD_LOCAL_HUB.instance = hub
    try:
        old_hub_coro = context_get('hub')
        context_set('hub', hub)
        await hub.open()
        try:
            yield
        finally:
            context_clear()
            await hub.close()
            context_set('hub', old_hub_coro)
    finally:
        __THREAD_LOCAL_HUB.instance = old_hub


def __get_root_hub() -> Optional[Hub]:
    return __ROOT_HUB


def __get_thread_local_gub() -> Optional[Hub]:
    # For now thread local context not used but code prepared for future purposes
    try:
        inst = __THREAD_LOCAL_HUB.instance
    except AttributeError:
        inst = None
    return inst


def _current_hub() -> Hub:
    inst = context_get('hub')
    if inst is None:
        root_hub = __get_thread_local_gub() or __get_root_hub()
        if root_hub is None:
            raise SiriusInitializationError('Non initialized Sirius Agent connection')
        inst = root_hub.copy()
        context_set('hub', inst)
    return inst



from abc import ABC, abstractmethod
from typing import Optional, List, Union, Tuple, Any

from sirius_sdk.agent.dkms import DKMS
from sirius_sdk.messaging import Message

from .bus import AbstractBus
from .p2p import Endpoint, Pairwise
from .listener import AbstractListener
from .batching import RoutingBatch


class APICoProtocols(ABC):

    @abstractmethod
    async def spawn_coprotocol(self) -> AbstractBus:
        raise NotImplemented


class APINetworks(ABC):

    @abstractmethod
    def dkms(self, name: str) -> Optional[DKMS]:
        raise NotImplemented


class APIRouter(ABC):

    @abstractmethod
    async def get_endpoints(self) -> List[Endpoint]:
        raise NotImplemented

    @abstractmethod
    async def subscribe(self, group_id: str = None) -> AbstractListener:
        raise NotImplemented


class APIDistributedLocks(ABC):

    @abstractmethod
    async def acquire(self, resources: List[str], lock_timeout: float, enter_timeout: float = None) -> (bool, List[str]):
        raise NotImplemented

    @abstractmethod
    async def release(self):
        raise NotImplemented


class APITransport(ABC):

    @abstractmethod
    async def send(
            self, message: Message, their_vk: Union[List[str], str],
            endpoint: str, my_vk: Optional[str] = None, routing_keys: Optional[List[str]] = None
    ):
        raise NotImplemented

    @abstractmethod
    async def send_to(self, message: Message, to: Pairwise):
        raise NotImplemented

    @abstractmethod
    async def send_batched(self, message: Message, batches: List[RoutingBatch]) -> List[Tuple[bool, str]]:
        raise NotImplemented


class APIContents(ABC):

    @abstractmethod
    async def generate_qr_code(self, value: str) -> str:
        raise NotImplemented


class APICrypto(ABC):

    @abstractmethod
    async def create_key(self, seed: str=None, crypto_type: str=None) -> str:
        """
        Creates keys pair and stores in the wallet.

        :param seed: string, (optional) Seed that allows deterministic key creation (if not set random one will be
                        created). Can be UTF-8, base64 or hex string.
        :param crypto_type: string, // Optional (if not set then ed25519 curve is used);
                        Currently only 'ed25519' value is supported for this field.

        :return: verkey: Ver key of generated key pair, also used as key identifier
        """
        raise NotImplemented()

    @abstractmethod
    async def set_key_metadata(self, verkey: str, metadata: dict) -> None:
        """
        Saves/replaces the meta information for the giving key in the wallet.

        :param verkey: the key (verkey, key id) to store metadata.
        :param metadata: the meta information that will be store with the key.
        """
        raise NotImplemented()

    @abstractmethod
    async def get_key_metadata(self, verkey: str) -> Optional[dict]:
        """
        Retrieves the meta information for the giving key in the wallet.

        :param verkey: The key (verkey, key id) to retrieve metadata.
        :return: metadata: The meta information stored with the key; Can be null if no metadata was saved for this key.
        """
        raise NotImplemented()

    @abstractmethod
    async def crypto_sign(self, signer_vk: str, msg: bytes) -> bytes:
        """
        Signs a message with a key.

        Note to use DID keys with this function you can call indy_key_for_did to get key id (verkey) for specific DID.

        :param signer_vk:  id (verkey) of my key. The key must be created by calling create_key or create_and_store_my_did
        :param msg: a message to be signed
        :return: a signature string
        """
        raise NotImplemented()

    @abstractmethod
    async def crypto_verify(self, signer_vk: str, msg: bytes, signature: bytes) -> bool:
        """
        Verify a signature with a verkey.

        Note to use DID keys with this function you can call key_for_did to get key id (verkey) for specific DID.

        :param signer_vk: verkey of signer of the message
        :param msg: message that has been signed
        :param signature: a signature to be verified
        :return: valid: true - if signature is valid, false - otherwise
        """
        raise NotImplemented()

    @abstractmethod
    async def anon_crypt(self, recipient_vk: str, msg: bytes) -> bytes:
        """
        Encrypts a message by anonymous-encryption scheme.

        Sealed boxes are designed to anonymously send messages to a Recipient given its public key.
        Only the Recipient can decrypt these messages, using its private key.
        While the Recipient can verify the integrity of the message, it cannot verify the identity of the Sender.

        Note to use DID keys with this function you can call key_for_did to get key id (verkey)
        for specific DID.

        Note: use pack_message function for A2A goals.

        :param recipient_vk: verkey of message recipient
        :param msg: a message to be signed
        :return: an encrypted message as an array of bytes
        """
        raise NotImplemented()

    @abstractmethod
    async def anon_decrypt(self, recipient_vk: str, encrypted_msg: bytes) -> bytes:
        """
        Decrypts a message by anonymous-encryption scheme.

        Sealed boxes are designed to anonymously send messages to a Recipient given its public key.
        Only the Recipient can decrypt these messages, using its private key.
        While the Recipient can verify the integrity of the message, it cannot verify the identity of the Sender.

        Note to use DID keys with this function you can call key_for_did to get key id (verkey)
        for specific DID.

        Note: use unpack_message function for A2A goals.

        :param recipient_vk: id (verkey) of my key. The key must be created by calling indy_create_key or create_and_store_my_did
        :param encrypted_msg: encrypted message
        :return: decrypted message as an array of bytes
        """
        raise NotImplemented()

    @abstractmethod
    async def pack_message(self, message: Any, recipient_verkeys: list, sender_verkey: str=None) -> bytes:
        """
        Packs a message by encrypting the message and serializes it in a JWE-like format (Experimental)

        Note to use DID keys with this function you can call did.key_for_did to get key id (verkey)
        for specific DID.

        :param message: the message being sent as a string. If it's JSON formatted it should be converted to a string
        :param recipient_verkeys: a list of Strings which are recipient verkeys
        :param sender_verkey: the sender's verkey as a string. -> When None is passed in this parameter, anoncrypt mode is used
        :returns an Agent Wire Message format as a byte array.
        """
        raise NotImplemented()

    @abstractmethod
    async def unpack_message(self, jwe: bytes) -> dict:
        """
        Unpacks a JWE-like formatted message outputted by pack_message (Experimental)

        #Returns:
        (Authcrypt mode)

        {
            "message": <decrypted message>,
            "recipient_verkey": <recipient verkey used to decrypt>,
            "sender_verkey": <sender verkey used to encrypt>
        }

        (Anoncrypt mode)

        {
            "message": <decrypted message>,
            "recipient_verkey": <recipient verkey used to decrypt>,
        }
        """
        raise NotImplemented()



from typing import Union, List, Optional


class RoutingBatch(dict):

    def __init__(
            self, their_vk: Union[List[str], str], endpoint: str,
            my_vk: Optional[str] = None, routing_keys: Optional[List[str]] = None, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        if isinstance(their_vk, str):
            self['recipient_verkeys'] = [their_vk]
        else:
            self['recipient_verkeys'] = their_vk
        self['endpoint_address'] = endpoint
        self['sender_verkey'] = my_vk
        self['routing_keys'] = routing_keys or []



from sirius_sdk.agent.aries_rfc.base import AriesProtocolMessage, RegisterMessage, THREAD_DECORATOR, \
    VALID_DOC_URI, AriesProblemReport
from typing import List, Optional, Dict


class CoProtocolMessage(AriesProtocolMessage, metaclass=RegisterMessage):
    """Aries feature 0482 Messages implementation

    https://github.com/hyperledger/aries-rfcs/tree/main/features/0482-coprotocol-protocol
    """
    DOC_URI = VALID_DOC_URI[0]
    PROTOCOL = 'coprotocol'

    def __init__(self, thid: str = None, pthid: str = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if thid or pthid:
            # https://github.com/hyperledger/aries-rfcs/tree/main/concepts/0011-decorators#basic-conventions
            threads = self.get(THREAD_DECORATOR, {})
            if thid:
                threads['thid'] = thid
            if pthid:
                threads['pthid'] = pthid
            self[THREAD_DECORATOR] = threads
        #
        for fld, value in kwargs.items():
            if fld not in ['id_', 'version', 'doc_uri']:
                self[fld] = value

    @property
    def thid(self) -> Optional[str]:
        return self.get(THREAD_DECORATOR, {}).get('thid', None)

    @thid.setter
    def thid(self, value: str):
        threads = self.get(THREAD_DECORATOR, {})
        threads['thid'] = value
        self[THREAD_DECORATOR] = threads

    @property
    def pthid(self) -> Optional[str]:
        return self.get(THREAD_DECORATOR, {}).get('pthid', None)


class CoProtocolBind(CoProtocolMessage, metaclass=RegisterMessage):
    NAME = 'bind'

    def __init__(
            self, thid: str = None, pthid: str = None, co_binding_id: str = None,
            cast: List[Dict] = None, *args, **kwargs
    ):
        super().__init__(thid=thid, pthid=pthid, *args, **kwargs)
        self['co_binding_id'] = co_binding_id
        if cast:
            self['cast'] = cast


class CoProtocolAttach(CoProtocolMessage, metaclass=RegisterMessage):
    NAME = 'attach'


class CoProtocolInput(CoProtocolMessage, metaclass=RegisterMessage):
    NAME = 'input'


class CoProtocolOutput(CoProtocolMessage, metaclass=RegisterMessage):
    NAME = 'output'


class CoProtocolDetach(CoProtocolMessage, metaclass=RegisterMessage):
    NAME = 'detach'


class CoProtocolProblemReport(AriesProblemReport, metaclass=RegisterMessage):
    PROTOCOL = CoProtocolMessage.PROTOCOL



from typing import List

import sirius_sdk


async def schedule_vaults(p2p: sirius_sdk.Pairwise, vaults: List[sirius_sdk.aries_rfc.EncryptedDataVault]):
    co = await sirius_sdk.spawn_coprotocol()
    await co.subscribe_ext(
        sender_vk=[p2p.their.verkey],
        recipient_vk=[p2p.me.verkey],
        protocols=[sirius_sdk.aries_rfc.BaseConfidentialStorageMessage.PROTOCOL]
    )
    protocol = sirius_sdk.aries_rfc.CalledEncryptedDataVault(caller=p2p, proxy_to=vaults)
    try:
        while True:
            e = await co.get_message()
            await protocol.handle(e.message)
    finally:
        await co.abort()



from sirius_sdk.agent.wallet.abstract.cache import CacheOptions, PurgeOptions
from sirius_sdk.agent.wallet.abstract.non_secrets import RetrieveRecordOptions
from sirius_sdk.agent.wallet.abstract.ledger import NYMRole, PoolAction
from sirius_sdk.agent.wallet.abstract import KeyDerivationMethod


__all__ = ["CacheOptions", "PurgeOptions", "RetrieveRecordOptions", "NYMRole", "PoolAction", "KeyDerivationMethod"]



import base64
import logging
import uuid
import contextlib

from sirius_sdk.encryption.ed25519 import ensure_is_bytes
from sirius_sdk.base import AbstractStateMachine, PersistentMixin
from sirius_sdk.abstract.p2p import Pairwise
from sirius_sdk.hub.coprotocols import CoProtocolThreadedP2P
from sirius_sdk.errors.exceptions import StateMachineAborted, OperationAbortedManually, StateMachineTerminatedWithError

from .components import EncryptedDataVault, DataVaultStreamWrapper
from .messages import *
from .streams import AbstractReadOnlyStream, AbstractWriteOnlyStream, BaseStreamEncryption, StreamEncryption, \
    StreamDecryption, ReadOnlyStreamDecodingWrapper, WriteOnlyStreamEncodingWrapper
from . import ConfidentialStorageEncType, EncryptedDocument, StructuredDocument, DocumentMeta, StreamMeta
from .errors import *
from .encoding import JWE, KeyPair
from .utils import *


def problem_report_from_exception(e: BaseConfidentialStorageError) -> ConfidentialStorageMessageProblemReport:
    problem_code, explain = e.PROBLEM_CODE, e.message
    report = ConfidentialStorageMessageProblemReport(
        problem_code=problem_code, explain=explain
    )
    return report


def exception_from_problem_report(
        report: ConfidentialStorageMessageProblemReport
) -> Optional[Union[BaseConfidentialStorageError, StateMachineTerminatedWithError]]:
    if report.problem_code == ConfidentialStorageUnexpectedMessageType.PROBLEM_CODE:
        return ConfidentialStorageUnexpectedMessageType(report.explain)
    elif report.problem_code == ConfidentialStorageInvalidRequest.PROBLEM_CODE:
        return ConfidentialStorageInvalidRequest(report.explain)
    elif report.problem_code == StreamEOF.PROBLEM_CODE:
        return StreamEOF(report.explain)
    elif report.problem_code == EncryptionError.PROBLEM_CODE:
        return EncryptionError(report.explain)
    elif report.problem_code == StreamInitializationError.PROBLEM_CODE:
        return StreamInitializationError(report.explain)
    elif report.problem_code == StreamSeekableError.PROBLEM_CODE:
        return StreamSeekableError(report.explain)
    elif report.problem_code == StreamFormatError.PROBLEM_CODE:
        return StreamFormatError(report.explain)
    elif report.problem_code == DocumentFormatError.PROBLEM_CODE:
        return DocumentFormatError(report.explain)
    elif report.problem_code == ConfidentialStorageTimeoutOccurred.PROBLEM_CODE:
        return ConfidentialStorageTimeoutOccurred(report.explain)
    elif report.problem_code == ConfidentialStoragePermissionDenied.PROBLEM_CODE:
        return ConfidentialStoragePermissionDenied(report.explain)
    elif report.problem_code == DataVaultCreateResourceError.PROBLEM_CODE:
        return DataVaultCreateResourceError(report.explain)
    elif report.problem_code == DataVaultResourceMissing.PROBLEM_CODE:
        return DataVaultResourceMissing(report.explain)
    elif report.problem_code == DataVaultSessionError.PROBLEM_CODE:
        return DataVaultSessionError(report.explain)
    elif report.problem_code == DataVaultStateError.PROBLEM_CODE:
        return DataVaultStateError(report.explain)
    elif report.problem_code == DataVaultOSError.PROBLEM_CODE:
        return DataVaultOSError(report.explain)
    else:
        return BaseConfidentialStorageError(report.explain)


class CallerReadOnlyStreamProtocol(AbstractStateMachine, AbstractReadOnlyStream):
    """ReadOnly Stream protocol for Caller entity

    See details:
        - Concept: https://github.com/hyperledger/aries-rfcs/blob/main/concepts/0478-coprotocols
    """

    def __init__(
            self, called: Pairwise, uri: str, read_timeout: int, retry_count: int = 1,
            thid: str = None, pthid: str = None, enc: BaseStreamEncryption = None, logger=None, *args, **kwargs
    ):
        """Stream abstraction for read-only operations provided with [Vault] entity

        :param called (required): Called entity who must process requests (Vault/Storage/etc)
        :param uri (required): address of stream resource
        :param read_timeout (required): time till operation is timeout occurred
        :param thid (optional): co-protocol thread-id
        :param pthid (optional): co-protocol parent-thread-id
        :param enc (optional): allow decrypt stream chunks
        :param retry_count (optional): if chunk-read-operation was terminated with timeout
                                       then protocol will re-try operation from the same seek
        :param logger (optional): state-machine logger
        """
        AbstractStateMachine.__init__(self, time_to_live=read_timeout, logger=logger, *args, **kwargs)
        AbstractReadOnlyStream.__init__(self, path=uri, chunks_num=0, enc=enc)
        self._problem_report: Optional[ConfidentialStorageMessageProblemReport] = None
        self.__uri = uri
        self.__called = called
        self.__retry_count = retry_count
        self.__thid = thid or uuid.uuid4().hex
        self.__pthid = pthid
        self.__coprotocol: Optional[CoProtocolThreadedP2P] = None

    @property
    def called(self) -> Pairwise:
        return self.__called

    @property
    def thid(self) -> str:
        return self.__thid

    @property
    def pthid(self) -> Optional[str]:
        return self.__pthid

    @property
    def read_timeout(self) -> int:
        return self.time_to_live

    @property
    def retry_count(self) -> int:
        return self.__retry_count

    async def open(self):
        resp = await self.rpc(
            request=StreamOperation(
                operation=StreamOperation.OperationCode.OPEN,
                params={
                    'uri': self.__uri,
                    'encrypted': self.enc is not None
                }
            )
        )
        state = resp.params['state']
        self._seekable = state.get('seekable', None)
        self._current_chunk = state.get('current_chunk', 0)
        self._chunks_num: int = state.get('chunks_num', 0)
        self._is_open = True

    async def close(self):
        if self._is_open:
            async with self.coprotocol() as co:
                self._is_open = False
                req = StreamOperation(
                    operation=StreamOperation.OperationCode.CLOSE
                )
                await co.send(req)
                self._current_chunk = 0
                self._chunks_num = 0
                self._seekable = None
            await self.close_coprotocol()

    async def seek_to_chunk(self, no: int) -> int:
        resp = await self.rpc(
            request=StreamOperation(
                operation=StreamOperation.OperationCode.SEEK_TO_CHUNK,
                params={
                    'no': no
                }
            )
        )
        no = resp.params['no']
        self._current_chunk = no
        return no

    async def read_chunk(self, no: int = None) -> (int, bytes):
        if no is not None:
            before_no = no
        else:
            before_no = self._current_chunk
        no, encrypted = await self.__internal_read_chunk(before_no)
        chunk = await self.decrypt(encrypted)
        return no, chunk

    async def eof(self) -> bool:
        return self._current_chunk >= self._chunks_num

    async def open_coprotocol(self) -> CoProtocolThreadedP2P:
        if self.__coprotocol is None:
            self.__coprotocol = CoProtocolThreadedP2P(
                thid=self.thid,
                pthid=self.pthid,
                to=self.called,
                time_to_live=self.read_timeout
            )
            self._register_for_aborting(self.__coprotocol)
        return self.__coprotocol

    async def close_coprotocol(self):
        if self.__coprotocol:
            await self.__coprotocol.clean()
            self._unregister_for_aborting(self.__coprotocol)
            self.__coprotocol = None

    @contextlib.asynccontextmanager
    async def coprotocol(self):
        co = await self.open_coprotocol()
        try:
            yield co
        except OperationAbortedManually:
            await self.log(progress=100, message='Aborted')
            raise StateMachineAborted('Aborted by User')

    async def rpc(self, request: StreamOperation) -> StreamOperationResult:
        async with self.coprotocol() as co:
            while True:
                success, resp = await co.switch(request)
                if success:
                    if isinstance(resp, StreamOperationResult):
                        return resp
                    elif isinstance(resp, ConfidentialStorageMessageProblemReport):
                        self._problem_report = resp
                        exc = exception_from_problem_report(resp)
                        raise exc
                    else:
                        # Co-Protocol will terminate if timeout occurred
                        pass
                else:
                    err_message = 'Caller read operation timeout occurred!'
                    report = ConfidentialStorageMessageProblemReport(
                        problem_code=ConfidentialStorageTimeoutOccurred.PROBLEM_CODE,
                        explain=err_message
                    )
                    await co.send(report)
                    raise ConfidentialStorageTimeoutOccurred(err_message)

    async def __internal_read_chunk(self, read_chunk: int) -> (int, bytes):
        before_no = read_chunk
        after_no = read_chunk + 1
        for n in range(self.retry_count):
            try:
                resp = await self.rpc(
                    request=StreamOperation(
                        operation=StreamOperation.OperationCode.READ_CHUNK,
                        params={
                            'no': before_no
                        }
                    )
                )
            except ConfidentialStorageTimeoutOccurred:
                # Next iteration
                await self.close_coprotocol()
                continue
            no = resp.params['no']
            if no == after_no:
                chunk = base64.b64decode(resp.params['chunk'])
                self._current_chunk = no
                return no, chunk
            else:
                # re-try again
                pass
        raise ConfidentialStorageTimeoutOccurred(
            f'Stream read Timeout occurred for timeout={self.read_timeout} and retry_count={self.retry_count}'
        )


class CalledReadOnlyStreamProtocol(PersistentMixin, AbstractStateMachine):
    """ReadOnly Stream protocol for Called entity, Called entity most probably operate on Vault side
       Acts as proxy to physical stream implementation

    See details:
        - Concept: https://github.com/hyperledger/aries-rfcs/blob/main/concepts/0478-coprotocols
    """

    def __init__(
            self, caller: Pairwise, thid: str = None, proxy_to: AbstractReadOnlyStream = None,
            persistent_id: str = None, time_to_live: int = None, logger=None, *args, **kwargs
    ):
        """Setup stream proxy environment

        :param caller: Caller of stream resource operations
        :param thid (optional): co-protocol thread-id
        :param proxy_to (optional): stream to proxy op calls
        :param persistent_id (optional): unique ID to store state to recover states on failures and restarts
        :param time_to_live (optional): time the state-machine is alive,
        :param logger (optional): state-machine logger
        """
        super(CalledReadOnlyStreamProtocol, self).__init__(time_to_live=time_to_live, logger=logger, *args, **kwargs)
        self.__caller: Pairwise = caller
        self.__thid = thid
        self.__proxy_to = proxy_to
        self.__persistent_id = persistent_id
        self.__persist_state = {}
        self.__storage_rec_exists = False
        self.__coprotocol: Optional[CoProtocolThreadedP2P] = None
        self._problem_report: Optional[ConfidentialStorageMessageProblemReport] = None

    @property
    def caller(self) -> Pairwise:
        return self.__caller

    @property
    def thid(self) -> str:
        return self.__thid

    @property
    def proxy_to(self) -> Optional[AbstractReadOnlyStream]:
        return self.__proxy_to

    @property
    def stream_is_open(self) -> bool:
        return self.__persist_state.get('is_open', False)

    @stream_is_open.setter
    def stream_is_open(self, value: bool):
        self.__persist_state['is_open'] = value

    async def run_forever(self, proxy_to: AbstractReadOnlyStream = None, exit_on_close: bool = True):
        """Proxy requests in loop

        :param proxy_to: stream interface implementation
        :param exit_on_close: (bool) exit loop on close
        :return:
        """
        if proxy_to is None:
            proxy_to = self.__proxy_to
        try:
            async with self.coprotocol(close_on_exit=True) as co:
                while True:
                    request, sender_verkey, recipient_verkey = await co.get_one()
                    await self.handle(request, proxy_to)
                    if isinstance(request, StreamOperation):
                        if request.operation == StreamOperation.OperationCode.CLOSE and exit_on_close:
                            return
        finally:
            if self.__coprotocol:
                await self.close_coprotocol()

    async def handle(
            self, request: Union[StreamOperation, ConfidentialStorageMessageProblemReport],
            proxy_to: AbstractReadOnlyStream = None
    ):
        if proxy_to is None:
            proxy_to = self.__proxy_to
        if isinstance(request, ConfidentialStorageMessageProblemReport):
            self._problem_report = request
        elif isinstance(request, StreamOperation):
            try:
                if proxy_to.is_open != self.stream_is_open:
                    if self.stream_is_open:
                        await proxy_to.open()
                    else:
                        await proxy_to.close()
                if request.operation == StreamOperation.OperationCode.OPEN:
                    encrypted = request.params.get('encrypted', False)
                    if encrypted:
                        # Local stream must persist chunks structure
                        if proxy_to.enc is None:
                            proxy_to.enc = BaseStreamEncryption(type_=ConfidentialStorageEncType.UNKNOWN)
                    await proxy_to.open()
                    params = {
                        'state': {
                            'seekable': proxy_to.seekable,
                            'chunks_num': proxy_to.chunks_num,
                            'current_chunk': proxy_to.current_chunk,
                        }
                    }
                    await self.__send_response(request, StreamOperationResult(request.operation, params))
                    self.stream_is_open = True
                elif request.operation == StreamOperation.OperationCode.CLOSE:
                    await proxy_to.close()
                    await self.__ensure_storage_record_missing()
                elif request.operation == StreamOperation.OperationCode.SEEK_TO_CHUNK:
                    no = request.params.get('no')
                    new_no = await proxy_to.seek_to_chunk(no)
                    params = {'no': new_no}
                    await self.__send_response(request, StreamOperationResult(request.operation, params))
                elif request.operation == StreamOperation.OperationCode.READ_CHUNK:
                    no = request.params.get('no')
                    new_no, chunk = await proxy_to.read_chunk(no)
                    chunk = base64.b64encode(chunk).decode()
                    params = {'no': new_no, 'chunk': chunk}
                    await self.__send_response(request, StreamOperationResult(request.operation, params))
            except BaseConfidentialStorageError as e:
                report = problem_report_from_exception(e)
                # Don't raise any error: give caller to make decision
                await self.__send_response(request, report)
                await self.__ensure_storage_record_missing()

    async def open_coprotocol(self) -> CoProtocolThreadedP2P:
        if self.__coprotocol is None:
            self.__coprotocol = CoProtocolThreadedP2P(
                thid=self.__thid,
                to=self.caller,
                time_to_live=self.time_to_live
            )
            self._register_for_aborting(self.__coprotocol)
        return self.__coprotocol

    async def close_coprotocol(self):
        if self.__coprotocol:
            await self.__coprotocol.clean()
            self._unregister_for_aborting(self.__coprotocol)
            self.__coprotocol = None
        await self.__ensure_storage_record_missing()

    async def abort(self):
        await self.__ensure_storage_record_missing()
        await super().abort()

    @contextlib.asynccontextmanager
    async def coprotocol(self, close_on_exit: bool = False):
        co = await self.open_coprotocol()
        try:
            try:
                yield co
            except OperationAbortedManually:
                await self.log(progress=100, message='Aborted')
                raise StateMachineAborted('Aborted by User')
        finally:
            if close_on_exit:
                await self.close_coprotocol()

    #############################
    # Persistent methods
    #############################
    async def load(self):
        if self.__proxy_to is not None and self.__persistent_id is not None:
            new_states = await self.__ensure_storage_record_exists()
            if new_states and new_states != self.__persist_state:
                self.__persist_state = new_states
                if self.__persist_state['is_open'] is False and self.__proxy_to.is_open is True:
                    await self.__proxy_to.close()
                if self.__persist_state['is_open'] is True and self.__proxy_to.is_open is False:
                    await self.__proxy_to.open()
                if self.__proxy_to.is_open and self.__persist_state['current_chunk'] != self.__proxy_to.current_chunk:
                    await self.__proxy_to.seek_to_chunk(self.__persist_state['current_chunk'])

    async def save(self):
        if self.__proxy_to is not None and self.__persistent_id is not None:
            self.__persist_state = {'is_open': self.__proxy_to.is_open, 'current_chunk': self.__proxy_to.current_chunk}
            await update_persist_record(self.__persistent_id, self.__persist_state)

    @property
    def edited(self) -> bool:
        if self.__proxy_to is not None:
            return self.__persist_state.get('is_open', None) != self.__proxy_to.is_open or \
                   self.__persist_state.get('current_chunk', None) != self.__proxy_to.current_chunk
        else:
            return False

    async def __send_response(
            self,
            request_: StreamOperation,
            response_: Union[StreamOperationResult, ConfidentialStorageMessageProblemReport]
    ):
        sirius_sdk.prepare_response(request_, response_)
        await sirius_sdk.send_to(response_, self.caller)

    async def __ensure_storage_record_exists(self) -> Optional[dict]:
        if self.__persistent_id is not None:
            states = await ensure_persist_record_exists(self.__persistent_id, self.__persist_state)
            self.__storage_rec_exists = True
            return states
        else:
            return None

    async def __ensure_storage_record_missing(self):
        if self.__storage_rec_exists:
            await ensure_persist_record_missing(self.__persistent_id)
            self.__storage_rec_exists = False


class CallerWriteOnlyStreamProtocol(AbstractStateMachine, AbstractWriteOnlyStream):
    """WriteOnly Stream protocol for Caller entity

    See details:
        - Concept: https://github.com/hyperledger/aries-rfcs/blob/main/concepts/0478-coprotocols
    """

    def __init__(
            self, called: Pairwise, uri: str, thid: str = None, pthid: str = None, enc: BaseStreamEncryption = None,
            retry_count: int = 3, time_to_live: int = 60, logger=None, *args, **kwargs
    ):
        """Stream abstraction for write-only operations provided with [Vault] entity

        :param called (required): Called entity who must process requests (Vault/Storage/etc)
        :param uri (required): address of stream resource
        :param thid (optional): co-protocol thread-id
        :param pthid (optional): co-protocol parent-thread-id
        :param enc: allow encrypt stream chunks
        :param retry_count (optional): if chunk-write-operation was terminated with timeout
                                       then protocol will re-try operation from the same seek
        :param logger (optional): state-machine logger
        """

        AbstractStateMachine.__init__(self, time_to_live=time_to_live, logger=logger, *args, **kwargs)
        AbstractWriteOnlyStream.__init__(self, path=uri, enc=enc)
        self._problem_report: Optional[ConfidentialStorageMessageProblemReport] = None
        self.__uri = uri
        self.__called = called
        self.__retry_count = retry_count
        self.__thid = thid or uuid.uuid4().hex
        self.__pthid = pthid
        self.__coprotocol: Optional[CoProtocolThreadedP2P] = None

    @property
    def called(self) -> Pairwise:
        return self.__called

    @property
    def thid(self) -> str:
        return self.__thid

    @property
    def pthid(self) -> Optional[str]:
        return self.__pthid

    @property
    def retry_count(self) -> int:
        return self.__retry_count

    async def open(self):
        resp = await self.rpc(
            request=StreamOperation(
                operation=StreamOperation.OperationCode.OPEN,
                params={
                    'uri': self.__uri,
                    'encrypted': self.enc is not None
                }
            )
        )
        state = resp.params['state']
        self._seekable = state.get('seekable', None)
        self._current_chunk = state.get('current_chunk', 0)
        self._chunks_num = state.get('chunks_num', 0)
        self.chunk_size = state.get('chunk_size', 0)
        self._is_open = True

    async def close(self):
        if self._is_open:
            async with self.coprotocol() as co:
                self._is_open = False
                req = StreamOperation(
                    operation=StreamOperation.OperationCode.CLOSE
                )
                await co.send(req)
                self._current_chunk = 0
                self._chunks_num = 0
                self._seekable = None
            await self.close_coprotocol()

    async def write_chunk(self, chunk: bytes, no: int = None) -> (int, int):
        if no is not None:
            before_no = no
        else:
            before_no = self._current_chunk
        encrypted = await self.encrypt(chunk)
        no, chunk = await self.__internal_write_chunk(before_no, encrypted)
        return no, chunk

    async def seek_to_chunk(self, no: int) -> int:
        resp = await self.rpc(
            request=StreamOperation(
                operation=StreamOperation.OperationCode.SEEK_TO_CHUNK,
                params={
                    'no': no
                }
            )
        )
        no = resp.params['no']
        self._current_chunk = no
        return no

    async def truncate(self, no: int = 0):
        resp = await self.rpc(
            request=StreamOperation(
                operation=StreamOperation.OperationCode.TRUNCATE,
                params={
                    'no': no
                }
            )
        )
        state = resp.params['state']
        self._current_chunk = state.get('current_chunk', self.current_chunk)
        self._chunks_num = state.get('chunks_num', self.chunks_num)

    async def open_coprotocol(self) -> CoProtocolThreadedP2P:
        if self.__coprotocol is None:
            self.__coprotocol = CoProtocolThreadedP2P(
                thid=self.thid,
                pthid=self.pthid,
                to=self.called,
                time_to_live=self.time_to_live
            )
            self._register_for_aborting(self.__coprotocol)
        return self.__coprotocol

    async def close_coprotocol(self):
        if self.__coprotocol:
            await self.__coprotocol.clean()
            self._unregister_for_aborting(self.__coprotocol)
            self.__coprotocol = None

    @contextlib.asynccontextmanager
    async def coprotocol(self):
        co = await self.open_coprotocol()
        try:
            yield co
        except OperationAbortedManually:
            await self.log(progress=100, message='Aborted')
            raise StateMachineAborted('Aborted by User')

    async def rpc(self, request: StreamOperation) -> StreamOperationResult:
        async with self.coprotocol() as co:
            while True:
                success, resp = await co.switch(request)
                if success:
                    if isinstance(resp, StreamOperationResult):
                        return resp
                    elif isinstance(resp, ConfidentialStorageMessageProblemReport):
                        self._problem_report = resp
                        exc = exception_from_problem_report(resp)
                        raise exc
                    else:
                        # Co-Protocol will terminate if timeout occurred
                        pass
                else:
                    err_message = 'Caller write operation timeout occurred!'
                    report = ConfidentialStorageMessageProblemReport(
                        problem_code=ConfidentialStorageTimeoutOccurred.PROBLEM_CODE,
                        explain=err_message
                    )
                    await co.send(report)
                    raise ConfidentialStorageTimeoutOccurred(err_message)

    async def __internal_write_chunk(self, chunk_no: int, chunk: bytes) -> (int, int):
        before_no = chunk_no
        chunk_b64 = base64.b64encode(chunk).decode()
        for n in range(self.retry_count):
            try:
                resp = await self.rpc(
                    request=StreamOperation(
                        operation=StreamOperation.OperationCode.WRITE_CHUNK,
                        params={
                            'no': before_no,
                            'chunk': chunk_b64
                        }
                    )
                )
            except ConfidentialStorageTimeoutOccurred:
                # Next iteration
                await self.close_coprotocol()
                continue
            new_offset = resp.params['no']
            writen_sz = resp.params['size']
            self._current_chunk = new_offset
            return new_offset, writen_sz
        raise ConfidentialStorageTimeoutOccurred(
            f'Stream write Timeout occurred for timeout={self.time_to_live} and retry_count={self.retry_count}'
        )


class CalledWriteOnlyStreamProtocol(PersistentMixin, AbstractStateMachine):
    """
        Called entity on WriteOnly Stream side, Called entity most probably operate on Vault side
        Acts as proxy to physical stream implementation

    See details:
        - Concept: https://github.com/hyperledger/aries-rfcs/blob/main/concepts/0478-coprotocols
    """

    def __init__(
            self, caller: Pairwise, thid: str = None, time_to_live: int = None,
            proxy_to: AbstractWriteOnlyStream = None, persistent_id: str = None,
            logger=None, *args, **kwargs
    ):
        """Setup stream proxy environment

        :param caller: Caller of stream resource operations
        :param thid (optional): co-protocol thread-id
        :param time_to_live (optional): time the state-machine is alive,
        :param proxy_to(optional): stream to proxy operations
        :param persistent_id (optional): unique ID to store state to recover states on failures and restarts
        :param logger (optional): state-machine logger
        """
        super().__init__(time_to_live=time_to_live, logger=logger, *args, **kwargs)
        self.__caller: Pairwise = caller
        self.__thid = thid
        self.__proxy_to = proxy_to
        self.__persistent_id = persistent_id
        self.__storage_rec_exists = False
        self.__persist_state = {}
        self.__coprotocol: Optional[CoProtocolThreadedP2P] = None
        self._problem_report: Optional[ConfidentialStorageMessageProblemReport] = None

    @property
    def caller(self) -> Pairwise:
        return self.__caller

    @property
    def thid(self) -> str:
        return self.__thid

    @property
    def proxy_to(self) -> Optional[AbstractWriteOnlyStream]:
        return self.__proxy_to

    @property
    def stream_is_open(self) -> bool:
        return self.__persist_state.get('is_open', False)

    @stream_is_open.setter
    def stream_is_open(self, value: bool):
        self.__persist_state['is_open'] = value

    async def run_forever(self, proxy_to: AbstractWriteOnlyStream = None, exit_on_close: bool = True):
        """Proxy requests in loop

        :param proxy_to: stream interface implementation
        :param exit_on_close: (bool) exit loop on close
        :return:
        """
        if proxy_to is None:
            proxy_to = self.__proxy_to
        try:
            async with self.coprotocol(close_on_exit=True) as co:
                while True:
                    request, sender_verkey, recipient_verkey = await co.get_one()
                    await self.handle(request, proxy_to)
                    if isinstance(request, StreamOperation):
                        if request.operation == StreamOperation.OperationCode.CLOSE and exit_on_close:
                            return
        finally:
            if self.__coprotocol:
                await self.close_coprotocol()

    async def handle(
            self, request: Union[StreamOperation, ConfidentialStorageMessageProblemReport],
            proxy_to: AbstractWriteOnlyStream = None
    ):
        if proxy_to is None:
            proxy_to = self.__proxy_to
        if isinstance(request, ConfidentialStorageMessageProblemReport):
            self._problem_report = request
        elif isinstance(request, StreamOperation):
            try:
                if proxy_to.is_open != self.stream_is_open:
                    if self.stream_is_open:
                        await proxy_to.open()
                    else:
                        await proxy_to.close()
                if request.operation == StreamOperation.OperationCode.OPEN:
                    encrypted = request.params.get('encrypted', False)
                    if encrypted:
                        # Local stream must persist chunks structure
                        if proxy_to.enc is None:
                            proxy_to.enc = BaseStreamEncryption(type_=ConfidentialStorageEncType.UNKNOWN)
                    await proxy_to.open()
                    params = {
                        'state': {
                            'seekable': proxy_to.seekable,
                            'chunks_num': proxy_to.chunks_num,
                            'current_chunk': proxy_to.current_chunk,
                            'chunk_size': proxy_to.chunk_size
                        }
                    }
                    await self.__send_response(request, StreamOperationResult(request.operation, params))
                    self.stream_is_open = True
                elif request.operation == StreamOperation.OperationCode.CLOSE:
                    await proxy_to.close()
                    await self.__ensure_storage_record_missing()
                elif request.operation == StreamOperation.OperationCode.SEEK_TO_CHUNK:
                    no = request.params.get('no')
                    new_no = await proxy_to.seek_to_chunk(no)
                    params = {'no': new_no}
                    await self.__send_response(request, StreamOperationResult(request.operation, params))
                elif request.operation == StreamOperation.OperationCode.WRITE_CHUNK:
                    no = request.params.get('no')
                    chunk = base64.b64decode(request.params.get('chunk'))
                    new_no, writen_sz = await proxy_to.write_chunk(chunk, no)
                    params = {'no': new_no, 'size': writen_sz}
                    await self.__send_response(request, StreamOperationResult(request.operation, params))
                elif request.operation == StreamOperation.OperationCode.TRUNCATE:
                    no = request.params.get('no')
                    await proxy_to.truncate(no)
                    params = {
                        'state': {
                            'chunks_num': proxy_to.chunks_num,
                            'current_chunk': proxy_to.current_chunk
                        }
                    }
                    await self.__send_response(request, StreamOperationResult(request.operation, params))
            except BaseConfidentialStorageError as e:
                report = problem_report_from_exception(e)
                # Don't raise any error: give caller to make decision
                await self.__send_response(request, report)
                await self.__ensure_storage_record_missing()

    async def open_coprotocol(self) -> CoProtocolThreadedP2P:
        if self.__coprotocol is None:
            self.__coprotocol = CoProtocolThreadedP2P(
                thid=self.__thid,
                to=self.caller,
                time_to_live=self.time_to_live
            )
            self._register_for_aborting(self.__coprotocol)
        return self.__coprotocol

    async def close_coprotocol(self):
        if self.__coprotocol:
            await self.__coprotocol.clean()
            self._unregister_for_aborting(self.__coprotocol)
            self.__coprotocol = None
        if self.__persistent_id:
            await self.__ensure_storage_record_missing()

    async def abort(self):
        await self.__ensure_storage_record_missing()
        await super().abort()

    @contextlib.asynccontextmanager
    async def coprotocol(self, close_on_exit: bool = False):
        co = await self.open_coprotocol()
        try:
            try:
                yield co
            except OperationAbortedManually:
                await self.log(progress=100, message='Aborted')
                raise StateMachineAborted('Aborted by User')
        finally:
            if close_on_exit:
                await self.close_coprotocol()

    #############################
    # Persistent methods
    #############################
    async def load(self):
        if self.__proxy_to is not None and self.__persistent_id is not None:
            new_states = await self.__ensure_storage_record_exists()
            if new_states and new_states != self.__persist_state:
                self.__persist_state = new_states
                if self.__persist_state['is_open'] is False and self.__proxy_to.is_open is True:
                    await self.__proxy_to.close()
                if self.__persist_state['is_open'] is True and self.__proxy_to.is_open is False:
                    await self.__proxy_to.open()
                if self.__proxy_to.is_open and self.__persist_state['current_chunk'] != self.__proxy_to.current_chunk:
                    await self.__proxy_to.seek_to_chunk(self.__persist_state['current_chunk'])

    async def save(self):
        if self.__proxy_to is not None and self.__persistent_id is not None:
            self.__persist_state = {'is_open': self.__proxy_to.is_open, 'current_chunk': self.__proxy_to.current_chunk}
            await update_persist_record(self.__persistent_id, self.__persist_state)

    @property
    def edited(self) -> bool:
        if self.__proxy_to is not None:
            return self.__persist_state.get('is_open', None) != self.__proxy_to.is_open or \
                   self.__persist_state.get('current_chunk', None) != self.__proxy_to.current_chunk
        else:
            return False

    async def __send_response(
            self,
            request_: StreamOperation,
            response_: Union[StreamOperationResult, ConfidentialStorageMessageProblemReport]
    ):
        sirius_sdk.prepare_response(request_, response_)
        await sirius_sdk.send_to(response_, self.caller)

    async def __ensure_storage_record_exists(self) -> Optional[dict]:
        if self.__persistent_id is not None:
            states = await ensure_persist_record_exists(self.__persistent_id, self.__persist_state)
            self.__storage_rec_exists = True
            return states
        else:
            return None

    async def __ensure_storage_record_missing(self):
        if self.__storage_rec_exists:
            await ensure_persist_record_missing(self.__persistent_id)
            self.__storage_rec_exists = False


class CallerEncryptedDataVault(AbstractStateMachine, EncryptedDataVault, EncryptedDataVault.Indexes):

    class CallerStreamWrapper(DataVaultStreamWrapper):

        def __init__(self, api: EncryptedDataVault, id_: str):
            self.__api = api
            self.__id = id_
            super().__init__(readable=None, writable=None)

        async def readable(self, jwe: Union[JWE, dict] = None, keys: KeyPair = None) -> AbstractReadOnlyStream:
            if self._readable is None:
                self._readable = await self.__api.readable(self.__id)
            if jwe is None:
                return self._readable
            else:
                enc = StreamDecryption.from_jwe(jwe)
                if keys is not None:
                    enc.setup(vk=keys.pk, sk=keys.sk)
                return ReadOnlyStreamDecodingWrapper(src=self._readable, enc=enc)

        async def writable(self, jwe: Union[JWE, dict] = None, cek: Union[bytes, str] = None) -> AbstractWriteOnlyStream:
            if self._writable is None:
                self._writable = await self.__api.writable(self.__id)
            if jwe is None:
                return self._writable
            else:
                if isinstance(cek, str):
                    cek = ensure_is_bytes(cek)
                return WriteOnlyStreamEncodingWrapper(dest=self._writable, enc=StreamEncryption.from_jwe(jwe, cek))

    def __init__(
            self, called: Pairwise, read_timeout: int = 30,
            retry_count: int = 3, logger=None, *args, **kwargs
    ):
        """Stream abstraction for write-only operations provided with [Vault] entity

        :param called (required): Called entity who must process requests
        :param uri (required): address of stream resource
        :param thid (optional): co-protocol thread-id
        :param read_timeout (optional): time till read-operation is timeout occurred
        :param retry_count (optional): if chunk-write-operation was terminated with timeout
                                       then protocol will re-try operation from the same seek
        :param logger (optional): state-machine logger
        """
        AbstractStateMachine.__init__(self, time_to_live=None, logger=logger, *args, **kwargs)
        EncryptedDataVault.__init__(self)
        self._problem_report: Optional[ConfidentialStorageMessageProblemReport] = None
        self.__thid: Optional[str] = None
        self.__coprotocol: Optional[CoProtocolThreadedP2P] = None
        self.__called = called
        self.__current_vault: Optional[str] = None
        self.__read_timeout = read_timeout
        self.__retry_count = retry_count

    @property
    def thid(self) -> Optional[str]:
        return self.__thid

    @property
    def is_open(self) -> bool:
        return self.__coprotocol is not None

    @property
    def called(self) -> Pairwise:
        return self.__called

    def select(self, vault: str):
        self.__current_vault = vault

    async def open(self):
        if self.__current_vault is None:
            raise RuntimeError('You should select vault (query list of ones at first)!')
        if self.is_open:
            raise RuntimeError('state machine already open! Close it before to start new session')
        session_id = uuid.uuid4().hex
        self.__thid = f'vault[{self.__current_vault}]/{session_id}'
        await self._open_coprotocol()
        await self._rpc(request=DataVaultOpen(vault=self.__current_vault))

    async def close(self):
        await self._rpc(request=DataVaultClose())
        await self._close_coprotocol()

    async def list_vaults(self) -> List[VaultConfig]:
        resp = await self._request(request=DataVaultQueryList())
        if isinstance(resp, DataVaultResponseList):
            return resp.vaults
        elif isinstance(resp, ConfidentialStorageMessageProblemReport):
            exc = exception_from_problem_report(resp)
            raise exc
        else:
            raise RuntimeError(f'Unexpected response type: {resp.__class__.__name__}')

    async def filter(self, **attributes) -> List[StructuredDocument]:
        resp = await self._rpc(
            request=DataVaultList(
                filters=dict(**attributes) if attributes else None
            )
        )
        if isinstance(resp, StructuredDocumentMessage):
            docs = []
            for attach in resp.documents:
                doc = self.__extract_structured_doc(attach)
                docs.append(doc)
            return docs
        else:
            raise ConfidentialStorageUnexpectedMessageType(resp)

    async def indexes(self) -> EncryptedDataVault.Indexes:
        return self

    async def create_stream(self, uri: str, meta: Union[dict, StreamMeta] = None, chunk_size: int = None, **attributes) -> StructuredDocument:
        resp = await self._rpc(
            request=DataVaultCreateStream(uri=uri, meta=meta, chunk_size=chunk_size, attributes=attributes)
        )
        if isinstance(resp, StructuredDocumentMessage):
            attach = resp.documents[0]
            doc = self.__extract_structured_doc(attach)
            return doc
        else:
            raise ConfidentialStorageUnexpectedMessageType(resp)

    async def create_document(self, uri: str, meta: Union[dict, DocumentMeta] = None, **attributes) -> StructuredDocument:
        resp = await self._rpc(
            request=DataVaultCreateDocument(uri=uri, meta=meta, attributes=attributes)
        )
        if isinstance(resp, StructuredDocumentMessage):
            attach = resp.documents[0]
            doc = self.__extract_structured_doc(attach)
            if doc.doc is None:
                doc.doc = EncryptedDocument(content=b'')
            return doc
        else:
            raise ConfidentialStorageUnexpectedMessageType(resp)

    async def remove(self, uri: str):
        await self._rpc(
            request=DataVaultRemoveResource(uri=uri)
        )

    async def update(self, uri: str, meta: Union[dict, StreamMeta, DocumentMeta] = None, **attributes):
        await self._rpc(
            request=DataVaultUpdateResource(uri=uri, meta=meta, attributes=attributes)
        )

    async def load(self, uri: str) -> StructuredDocument:
        resp = await self._rpc(
            request=DataVaultLoadResource(uri=uri)
        )
        if isinstance(resp, StructuredDocumentMessage):
            attach = resp.documents[0]
            doc = self.__extract_structured_doc(attach)
            return doc
        else:
            raise ConfidentialStorageUnexpectedMessageType(resp)

    async def save_document(self, uri: str, doc: EncryptedDocument):
        request = DataVaultSaveDocument(uri=uri)
        request.document = doc
        await self._rpc(request)

    async def readable(self, uri: str) -> AbstractReadOnlyStream:
        stream = CallerReadOnlyStreamProtocol(
            called=self.called, uri=uri, pthid=self.thid,
            read_timeout=self.__read_timeout,
            retry_count=self.__retry_count,
        )
        await self._rpc(
            request=DataVaultBindStreamForReading(uri=uri, co_binding_id=stream.thid)
        )
        return stream

    async def writable(self, uri: str) -> AbstractWriteOnlyStream:
        stream = CallerWriteOnlyStreamProtocol(
            called=self.called, uri=uri, pthid=self.thid,
            retry_count=self.__retry_count,
            time_to_live=60*60  # 1 hour
        )
        await self._rpc(
            request=DataVaultBindStreamForWriting(uri=uri, co_binding_id=stream.thid)
        )
        return stream

    async def _request(self, request: BaseDataVaultOperation) -> BaseConfidentialStorageMessage:
        co = sirius_sdk.CoProtocolP2P(
            pairwise=self.called,
            protocols=[BaseDataVaultOperation.PROTOCOL],
            time_to_live=15
        )
        await co.start()
        try:
            ok, resp = await co.switch(request)
            if ok:
                return resp
            else:
                err_message = 'Caller write operation timeout occurred!'
                report = ConfidentialStorageMessageProblemReport(
                    problem_code=ConfidentialStorageTimeoutOccurred.PROBLEM_CODE,
                    explain=err_message
                )
                await co.send(report)
                raise ConfidentialStorageTimeoutOccurred(err_message)
        finally:
            await co.stop()

    async def _rpc(self, request: BaseDataVaultOperation) -> Union[BaseDataVaultOperation, DataVaultOperationAck, StructuredDocumentMessage]:
        if not self.is_open:
            raise RuntimeError('Open Vault at first!')
        async with self._coprotocol() as co:
            while True:
                success, resp = await co.switch(request)
                if success:
                    if isinstance(resp, BaseDataVaultOperation) or isinstance(resp, DataVaultOperationAck) or isinstance(resp, StructuredDocumentMessage):
                        return resp
                    elif isinstance(resp, ConfidentialStorageMessageProblemReport):
                        self._problem_report = resp
                        exc = exception_from_problem_report(resp)
                        raise exc
                    else:
                        raise ConfidentialStorageUnexpectedMessageType(resp)
                else:
                    err_message = 'Caller write operation timeout occurred!'
                    raise ConfidentialStorageTimeoutOccurred(err_message)

    async def _open_coprotocol(self) -> CoProtocolThreadedP2P:
        if self.__coprotocol is None:
            self.__coprotocol = CoProtocolThreadedP2P(
                thid=self.__thid,
                to=self.__called,
                time_to_live=self.time_to_live
            )
            self._register_for_aborting(self.__coprotocol)
        return self.__coprotocol

    async def _close_coprotocol(self):
        if self.__coprotocol:
            await self.__coprotocol.clean()
            self._unregister_for_aborting(self.__coprotocol)
            self.__coprotocol = None

    @contextlib.asynccontextmanager
    async def _coprotocol(self, close_on_exit: bool = False):
        co = await self._open_coprotocol()
        try:
            try:
                yield co
            except OperationAbortedManually:
                await self.log(progress=100, message='Aborted')
                raise StateMachineAborted('Aborted by User')
        finally:
            if close_on_exit:
                await self._close_coprotocol()

    def __extract_structured_doc(self, attach: StructuredDocumentAttach) -> StructuredDocument:
        doc = StructuredDocument(
            id_=attach.id,
            urn=attach.urn,
            meta=attach.meta,
            indexed=[
                StructuredDocument.Index(
                    sequence=ind.sequence,
                    hmac=ind.hmac,
                    attributes=ind.attributes
                )
                for ind in attach.indexed
            ],
            content=attach.document,
        )
        if attach.stream is not None:
            doc.stream = self.CallerStreamWrapper(api=self, id_=attach.stream.id)
        return doc


class CalledEncryptedDataVault(AbstractStateMachine):

    STREAM_CATEGORY_READER = 'reader'
    STREAM_CATEGORY_WRITER = 'writer'

    def __init__(self, caller: Pairwise, proxy_to: List[EncryptedDataVault], logger=None, *args, **kwargs):
        """Setup proxy environment

        :param thid (optional): co-protocol thread-id
        :param logger (optional): state-machine logger
        """
        super().__init__(time_to_live=None, logger=logger, *args, **kwargs)
        self.__coprotocol: Optional[CoProtocolThreadedP2P] = None
        self.__proxy_to = proxy_to
        self.__caller = caller
        self.__sessions: Dict[str, EncryptedDataVault] = {}
        self.__read_sessions: Dict[str, Dict[str, CalledReadOnlyStreamProtocol]] = {}
        self.__write_sessions: Dict[str, Dict[str, CalledWriteOnlyStreamProtocol]] = {}
        self._problem_report: Optional[ConfidentialStorageMessageProblemReport] = None

    @property
    def caller(self) -> Pairwise:
        return self.__caller

    @property
    def proxy_to(self) -> List[EncryptedDataVault]:
        return self.__proxy_to

    async def handle(self, request: BaseConfidentialStorageMessage):
        try:
            if isinstance(request, ConfidentialStorageMessageProblemReport):
                # Process problem-report
                self._problem_report = request
                logging.warning(
                    f'Received problem report problem_code: "{request.problem_code}" explain: "{request.explain}"'
                )
            elif isinstance(request, DataVaultQueryList):
                # Process query of list of available Vaults
                response = DataVaultResponseList(vaults=[vault.cfg for vault in self.proxy_to])
                sirius_sdk.prepare_response(request, response)
                await sirius_sdk.send_to(response, self.caller)
            elif isinstance(request, BaseDataVaultOperation):
                # Process Vault operation
                thread = request.thread
                if thread is None or not thread.thid:
                    raise ConfidentialStorageInvalidRequest(
                        'Data Vault request should have "~thread.thid" attribute to bind sessions'
                    )
                if isinstance(request, DataVaultOpen):
                    # Request to open
                    if not request.vault:
                        raise ConfidentialStorageInvalidRequest('You should set "vault" attribute')
                    proxy_to = [
                        vault for vault in self.proxy_to
                        if vault.cfg.id == request.vault or vault.cfg.reference_id == request.vault
                    ]
                    if not proxy_to:
                        raise ConfidentialStorageInvalidRequest(f'Vault with ID "{request.vault}" dow bot exists')
                    vault: EncryptedDataVault = proxy_to[0]
                    await vault.open()
                    self.__sessions[request.thread.thid] = vault
                    await self.__send_response(request, DataVaultOperationAck())
                elif isinstance(request, DataVaultClose):
                    # Request to close
                    vault = self.__sessions.get(request.thread.thid, None)
                    if vault is not None:
                        read_sessions = self.__read_sessions.get(request.thread.thid, {})
                        for thid, proto in read_sessions.items():
                            await proto.proxy_to.close()
                            await proto.close_coprotocol()
                            # Clean storage
                            persist_id_uri, _ = self.__build_persist_ids(
                                thid=request.thread.thid, co_binding_id=thid, category=self.STREAM_CATEGORY_READER
                            )
                            await ensure_persist_record_missing(persist_id_uri)
                        if read_sessions:
                            del self.__read_sessions[request.thread.thid]
                        write_sessions = self.__write_sessions.get(request.thread.thid, {})
                        for thid, proto in write_sessions.items():
                            await proto.proxy_to.close()
                            await proto.close_coprotocol()
                            # Clean storage
                            persist_id_uri, _ = self.__build_persist_ids(
                                thid=request.thread.thid, co_binding_id=thid, category=self.STREAM_CATEGORY_WRITER
                            )
                            await ensure_persist_record_missing(persist_id_uri)
                        if write_sessions:
                            del self.__write_sessions[request.thread.thid]
                        await vault.close()
                        del self.__sessions[request.thread.thid]
                    await self.__send_response(request, DataVaultOperationAck())
                elif isinstance(request, DataVaultCreateStream):
                    vault = self.__get_vault(request.thread.thid)
                    if not request.uri:
                        raise ConfidentialStorageInvalidRequest('You should set "uri" attribute')
                    attributes = request.attributes or {}
                    structured_document = await vault.create_stream(
                        uri=request.uri, meta=request.meta, chunk_size=request.chunk_size, **attributes
                    )
                    response = StructuredDocumentMessage(
                        documents=[
                            StructuredDocumentAttach.create_from(src=structured_document, sequence=0)
                        ]
                    )
                    await self.__send_response(request, response)
                elif isinstance(request, DataVaultCreateDocument):
                    vault = self.__get_vault(request.thread.thid)
                    if not request.uri:
                        raise ConfidentialStorageInvalidRequest('You should set "uri" attribute')
                    attributes = request.attributes or {}
                    structured_document = await vault.create_document(
                        uri=request.uri, meta=request.meta, **attributes
                    )
                    response = StructuredDocumentMessage(
                        documents=[
                            StructuredDocumentAttach.create_from(src=structured_document, sequence=0)
                        ]
                    )
                    await self.__send_response(request, response)
                elif isinstance(request, DataVaultUpdateResource):
                    vault = self.__get_vault(request.thread.thid)
                    if not request.uri:
                        raise ConfidentialStorageInvalidRequest('You should set "uri" attribute')
                    attributes = request.attributes or {}
                    await vault.update(
                        uri=request.uri, meta=request.meta, **attributes
                    )
                    await self.__send_response(request, DataVaultOperationAck())
                elif isinstance(request, DataVaultLoadResource):
                    vault = self.__get_vault(request.thread.thid)
                    if not request.uri:
                        raise ConfidentialStorageInvalidRequest('You should set "uri" attribute')
                    structured_document = await vault.load(request.uri)
                    response = StructuredDocumentMessage(
                        documents=[
                            StructuredDocumentAttach.create_from(src=structured_document, sequence=0)
                        ]
                    )
                    await self.__send_response(request, response)
                elif isinstance(request, DataVaultRemoveResource):
                    vault = self.__get_vault(request.thread.thid)
                    if not request.uri:
                        raise ConfidentialStorageInvalidRequest('You should set "uri" attribute')
                    await vault.remove(request.uri)
                    await self.__send_response(request, DataVaultOperationAck())
                elif isinstance(request, DataVaultSaveDocument):
                    vault = self.__get_vault(request.thread.thid)
                    if not request.uri:
                        raise ConfidentialStorageInvalidRequest('You should set "uri" attribute')
                    if not request.document:
                        raise ConfidentialStorageInvalidRequest('You should attach document with "doc~attach"')
                    await vault.save_document(uri=request.uri, doc=request.document)
                    await self.__send_response(request, DataVaultOperationAck())
                elif isinstance(request, DataVaultList):
                    vault = self.__get_vault(request.thread.thid)
                    attributes = request.filters
                    indexes = await vault.indexes()
                    structured_documents = await indexes.filter(**attributes)
                    response = StructuredDocumentMessage(
                        documents=[
                            StructuredDocumentAttach.create_from(src=doc, sequence=i)
                            for i, doc in enumerate(structured_documents)
                        ]
                    )
                    await self.__send_response(request, response)
                elif isinstance(request, DataVaultBindStreamForWriting) or isinstance(request, DataVaultBindStreamForReading):
                    vault = self.__get_vault(request.thread.thid)
                    if not request.uri:
                        raise ConfidentialStorageInvalidRequest('You should set "uri" attribute')
                    if not request.co_binding_id:
                        raise ConfidentialStorageInvalidRequest('You should set "co_binding_id" attribute to bind thread')
                    if isinstance(request, DataVaultBindStreamForWriting):
                        category = self.STREAM_CATEGORY_WRITER
                        # stream = await vault.writable(uri=request.uri)
                        # proto = CalledWriteOnlyStreamProtocol(
                        #     caller=self.caller, thid=request.co_binding_id, proxy_to=stream
                        # )
                        # write_sessions = self.__write_sessions.get(thread.thid, {})
                        # write_sessions[request.co_binding_id] = proto
                        # self.__write_sessions[thread.thid] = write_sessions
                    else:
                        category = self.STREAM_CATEGORY_READER
                        # stream = await vault.readable(uri=request.uri)
                        # proto = CalledReadOnlyStreamProtocol(
                        #     caller=self.caller, thid=request.co_binding_id, proxy_to=stream
                        # )
                        # read_sessions = self.__read_sessions.get(thread.thid, {})
                        # read_sessions[request.co_binding_id] = proto
                        # self.__read_sessions[thread.thid] = read_sessions
                    persist_id_uri, _ = self.__build_persist_ids(
                        thid=request.thread.thid, co_binding_id=request.co_binding_id, category=category
                    )
                    await store_persist_record_value(persist_id_uri, request.uri)
                    await self.__send_response(request, DataVaultOperationAck())
                else:
                    raise ConfidentialStorageUnexpectedMessageType(request)
            elif isinstance(request, StreamOperation):
                # Process Stream operation
                thread = request.thread
                if thread is None or not thread.thid:
                    raise ConfidentialStorageInvalidRequest(
                        'Stream operation request should have "~thread.thid" attribute to bind sessions'
                    )
                if thread.pthid is None:
                    raise ConfidentialStorageInvalidRequest(
                        'Stream operation request should have "~thread.pthid" attribute to bind sessions'
                    )
                proto = await self.__load_stream(request.thread)
                if proto is not None:
                    await proto.handle(request)
                    if proto.edited:
                        await proto.save()
                # proto = self.__write_sessions.get(thread.pthid, {}).get(thread.thid, None)
                # if proto is not None:
                #     await proto.handle(request)
                # proto = self.__read_sessions.get(thread.pthid, {}).get(thread.thid, None)
                # if proto is not None:
                #     await proto.handle(request)
            else:
                raise ConfidentialStorageUnexpectedMessageType(request)
        except BaseConfidentialStorageError as e:
            report = problem_report_from_exception(e)
            # Don't raise any error: give caller to make decision
            await self.__send_response(request, report)
        except OSError as e:
            exc = DataVaultOSError(e)
            report = problem_report_from_exception(exc)
            await self.__send_response(request, report)

    def __get_vault(self, thid: str) -> EncryptedDataVault:
        vault = self.__sessions.get(thid, None)
        if vault is None:
            raise ConfidentialStorageInvalidRequest(f'Not found vault bind for session thread: {thid}')
        return vault

    async def __load_stream(
            self, thread: BaseDataVaultOperation.Thread,
    ) -> Optional[Union[CalledWriteOnlyStreamProtocol, CalledReadOnlyStreamProtocol]]:
        reading_proto = self.__read_sessions.get(thread.pthid, {}).get(thread.thid, None)
        if reading_proto is not None:
            return reading_proto
        writing_proto = self.__write_sessions.get(thread.pthid, {}).get(thread.thid, None)
        if writing_proto is not None:
            return writing_proto
        # If no one is found then try to load states
        vault = self.__get_vault(thread.pthid)
        for category in [self.STREAM_CATEGORY_READER, self.STREAM_CATEGORY_WRITER]:
            persist_id_uri, persist_id_proto = self.__build_persist_ids(
                thid=thread.pthid, co_binding_id=thread.thid, category=category
            )
            uri = await load_persist_record_value(persist_id_uri)
            if uri is not None:
                # Restore states
                if category == self.STREAM_CATEGORY_WRITER:
                    stream = await vault.writable(uri)
                    proto = CalledWriteOnlyStreamProtocol(
                        caller=self.caller, thid=thread.thid, proxy_to=stream, persistent_id=persist_id_proto
                    )
                    await proto.load()
                    write_sessions = self.__write_sessions.get(thread.pthid, {})
                    write_sessions[thread.thid] = proto
                    self.__write_sessions[thread.pthid] = write_sessions
                    return proto
                elif category == self.STREAM_CATEGORY_READER:
                    stream = await vault.readable(uri)
                    proto = CalledReadOnlyStreamProtocol(
                        caller=self.caller, thid=thread.thid, proxy_to=stream, persistent_id=persist_id_proto
                    )
                    await proto.load()
                    read_sessions = self.__read_sessions.get(thread.pthid, {})
                    read_sessions[thread.thid] = proto
                    self.__read_sessions[thread.pthid] = read_sessions
                    return proto
        return None

    @staticmethod
    def __build_persist_ids(thid: str, co_binding_id: str, category: str) -> (str, str):
        common_prefix = f'{thid}/{co_binding_id}/{category}'
        persist_id_for_uri = f'{common_prefix}:uri'
        persist_id_for_proto = f'{common_prefix}:proto'
        return persist_id_for_uri, persist_id_for_proto

    async def __send_response(
            self,
            request_: BaseConfidentialStorageMessage,
            response_: Union[BaseConfidentialStorageMessage, DataVaultOperationAck, ConfidentialStorageMessageProblemReport]
    ):
        sirius_sdk.prepare_response(request_, response_)
        await sirius_sdk.send_to(response_, self.caller)



import warnings
from typing import Optional, List, Union, Tuple

from sirius_sdk.abstract.bus import AbstractBus
from sirius_sdk.abstract.batching import RoutingBatch
from sirius_sdk.abstract.listener import AbstractListener
from sirius_sdk.agent.pairwise import AbstractPairwiseList
from sirius_sdk.abstract.api import APICrypto
from sirius_sdk.agent.wallet.abstract.non_secrets import AbstractNonSecrets
from sirius_sdk.agent.wallet.abstract.cache import AbstractCache
from sirius_sdk.agent.wallet.abstract.did import AbstractDID
from sirius_sdk.messaging import Message
from sirius_sdk.agent.dkms import Ledger, DKMS
from sirius_sdk.abstract.p2p import Endpoint, Pairwise
from sirius_sdk.agent.microledgers.abstract import AbstractMicroledgerList
from sirius_sdk.errors.exceptions import SiriusInitializationError

from .config import Config
from .core import _current_hub, init, context
from .proxies import DIDProxy, CryptoProxy, MicroledgersProxy, PairwiseProxy, AnonCredsProxy, \
    CacheProxy, NonSecretsProxy
from .coprotocols import CoProtocolThreadedP2P, CoProtocolP2PAnon, CoProtocolP2P, AbstractP2PCoProtocol, \
    CoProtocolThreadedTheirs, open_communication, prepare_response

DID: AbstractDID = DIDProxy()
Crypto: APICrypto = CryptoProxy()
Microledgers: AbstractMicroledgerList = MicroledgersProxy()
PairwiseList: AbstractPairwiseList = PairwiseProxy()
AnonCreds: AnonCredsProxy = AnonCredsProxy()
Cache: AbstractCache = CacheProxy()
NonSecrets: AbstractNonSecrets = NonSecretsProxy()


async def dkms(name: str) -> Optional[DKMS]:
    api = await _current_hub().get_networks()
    if api is None:
        raise SiriusInitializationError('Configure APINetworks via "Config.override_networks(...)"')
    return api.dkms(name)


async def ledger(name: str) -> Optional[Ledger]:
    """!!! Deprecated Call !!!"""
    warnings.warn('Use sirius_sdk.dkms instead of this call', DeprecationWarning)
    return await dkms(name)


async def spawn_coprotocol() -> AbstractBus:
    api = await _current_hub().get_coprotocols()
    if api is None:
        raise SiriusInitializationError('Configure APICoProtocols via "Config.override_coprotocols(...)"')
    bus = await api.spawn_coprotocol()
    return bus


async def endpoints() -> List[Endpoint]:
    api = await _current_hub().get_router()
    if api is None:
        raise SiriusInitializationError('Configure APIRouter via "Config.override_router(...)"')
    return await api.get_endpoints()


async def subscribe(group_id: str = None) -> AbstractListener:
    api = await _current_hub().get_router()
    if api is None:
        raise SiriusInitializationError('Configure APIRouter via "Config.override_router(...)"')
    return await api.subscribe(group_id=group_id)


async def ping() -> bool:
    success = await _current_hub().ping()
    return success


async def send(
        message: Message, their_vk: Union[List[str], str],
        endpoint: str, my_vk: Optional[str], routing_keys: Optional[List[str]] = None
):
    api = await _current_hub().get_transport()
    if api is None:
        raise SiriusInitializationError('Configure APITransport via "Config.override_transport(...)"')
    await api.send(
        message=message, their_vk=their_vk,
        endpoint=endpoint, my_vk=my_vk, routing_keys=routing_keys
    )


async def send_to(message: Message, to: Pairwise):
    api = await _current_hub().get_transport()
    if api is None:
        raise SiriusInitializationError('Configure APITransport via "Config.override_transport(...)"')
    await api.send_to(message=message, to=to)


async def send_batched(message: Message, batches: List[RoutingBatch]) -> List[Tuple[bool, str]]:
    api = await _current_hub().get_transport()
    if api is None:
        raise SiriusInitializationError('Configure APITransport via "Config.override_transport(...)"')
    results = await api.send_batched(message, batches)
    return results


async def generate_qr_code(value: str) -> str:
    api = await _current_hub().get_contents()
    if api is None:
        raise SiriusInitializationError('Configure APIContents via "Config.override_contents(...)"')
    url = await api.generate_qr_code(value)
    return url


async def acquire(resources: List[str], lock_timeout: float, enter_timeout: float = None) -> (bool, List[str]):
    api = await _current_hub().get_distr_locks()
    if api is None:
        raise SiriusInitializationError('Configure APIDistributedLocks via "Config.override_distr_locks(...)"')
    kwargs = {'resources': resources, 'lock_timeout': lock_timeout}
    if enter_timeout:
        kwargs['enter_timeout'] = enter_timeout
    return await api.acquire(**kwargs)


async def release():
    api = await _current_hub().get_distr_locks()
    if api is None:
        raise SiriusInitializationError('Configure APIDistributedLocks via "Config.override_distr_locks(...)"')
    await api.release()
